from __future__ import annotations

import functools
import typing as T

import torch
from torch import Tensor

__all__ = []


def bchw_to_blc(x: Tensor) -> tuple[Tensor, torch.Size]:
    """
    Convert a tensor from BCHW to BLC format.

    Parameters
    ----------
    x : Tensor[B, C, H, W]
        The input tensor, with dimensions: batch, channels, height, width.

    Returns
    -------
    Tensor[B, L, C]
        The ``reshape``d tensor, with dimensions: batch, length, channels. The length is height * width.
    torch.Size[H, W]
        The shape of the spatial dimensions of the input tensor.
    """
    spatial_shape = x.shape[2:]
    x = x.permute(0, 2, 3, 1).reshape(x.shape[0], -1, x.shape[1])

    return x.contiguous(), spatial_shape


def blc_to_bchw(x: Tensor, spatial_shape: torch.Size) -> Tensor:
    """
    Convert a tensor from BLC to BCHW format.

    Parameters
    ----------
    x : Tensor[B, L, C]
        The input tensor, with dimensions: batch, length, channels.
    spatial_shape : torch.Size[H, W]
        The shape of the spatial dimensions of the input tensor.

    Returns
    -------
    Tensor[B, C, H, W]
        The ``reshape``d tensor, with dimensions: batch, channels, height, width.
    """
    h, w = spatial_shape
    x = x.reshape(x.shape[0], h, w, x.shape[2]).permute(0, 3, 1, 2)

    return x.contiguous()


def with_bchw_as_blc(fn: T.Callable[[Tensor], Tensor]) -> T.Callable[[Tensor], Tensor]:
    """Wraps a function that operats on a BLC tensor to work with BCHW tensors."""

    @functools.wraps(fn)
    def wrapper(x: Tensor) -> Tensor:
        x, spatial_shape = bchw_to_blc(x)
        x = fn(x)
        return blc_to_bchw(x, spatial_shape)

    return wrapper


def cat_nonempty(tensors: T.Iterable[Tensor | None], dim=0) -> Tensor | None:
    """
    Concatenate an interable of tensors for each entry that has a length greater
    than zero.

    Parameters
    ----------
    tensors
        T.Iterable of `Tensor` objects.
    dim, optional
        Dimension to concatenate, by default 0

    Returns
    -------
        Concatenated tensor.
    """
    tensors = list(tensors)
    tensors_nonempty = [t for t in tensors if t is not None]
    if len(tensors_nonempty) == 0:
        return None
    return torch.cat(tensors_nonempty, dim=dim)


def topk_score(scores, *, K: int, score_shape: torch.Size):
    """
    Get the top $k$ points in a scope mat.

    Parameters
    ----------
    scores
        Map of scores.
    K
        Amount of points to select.
    score_shape
        Original shape of the score map.

    Returns
    -------
        Tensors of the Top-K (scores, indices, classes, y, x)
    """
    match len(score_shape):
        case 4:
            batch, channel, height, width = score_shape
        case 3:
            batch, channel, height, width = 1, *score_shape
        case _:
            msg = "Expected 3 or 4D tensor"
            raise ValueError(msg)

    # get topk score and its index in every H x W(channel dim) feature map
    topk_scores, topk_inds = torch.topk(scores.reshape(batch, channel, -1), K)

    topk_inds = topk_inds % (height * width)
    topk_ys = torch.div(topk_inds, width, rounding_mode="floor").float()
    topk_xs = (topk_inds % width).int().float()

    # get all topk in in a batch
    topk_score, index = torch.topk(topk_scores.reshape(batch, -1), K)
    # div by K because index is grouped by K(C x K shape)
    topk_clses = torch.div(index, K, rounding_mode="floor")
    topk_inds = gather_feature(topk_inds.view(batch, -1, 1), index).reshape(batch, K)
    topk_ys = gather_feature(topk_ys.reshape(batch, -1, 1), index).reshape(batch, K)
    topk_xs = gather_feature(topk_xs.reshape(batch, -1, 1), index).reshape(batch, K)

    return topk_score, topk_inds, topk_clses, topk_ys, topk_xs


def gather_feature(
    fmap: Tensor,
    index: Tensor,
    mask: Tensor | None = None,
    use_transform: bool = False,
) -> Tensor:
    """
    Gather features from a mapping.

    Parameters
    ----------
    fmap
        Feature map.
    index
        Index tensor.
    mask, optional
        Masking tensor, by default None
    use_transform, optional
        Flag to apply transformation, by default False

    Returns
    -------
        Tensor with gathered features.
    """
    if use_transform:
        # change a (N, C, H, W) tenor to (N, HxW, C) shape
        batch, channel = fmap.shape[:2]
        fmap = fmap.view(batch, channel, -1).permute((0, 2, 1)).contiguous()

    dim = fmap.size(-1)
    index = index.unsqueeze(len(index.shape)).expand(*index.shape, dim)
    fmap = fmap.gather(dim=1, index=index)
    if mask is not None:
        mask = mask.unsqueeze(2).expand_as(fmap)
        fmap = fmap[mask]
        fmap = fmap.reshape(-1, dim)
    return fmap


def map_values(
    x: Tensor,
    translation: Tensor | tuple[Tensor, Tensor] | dict[int, int | float | bool],
    default: float | bool | None = None,
) -> Tensor:
    """
    Maps the values in a tensor to a new set of values.

    Parameters
    ----------
    x
        Input tensor.
    translation
        Mapping from old values to new values.

    Returns
    -------
    Tensor
        Tensor with mapped values.


    Examples
    --------
    >>> t = (torch.arange(0, 256), torch.randperm(256))
    >>> x = torch.randint(0, 256, (16, 224, 224, 3))
    >>> y = map_values(x, t)
    """
    # Find the translation sources and target values
    if isinstance(translation, dict):
        assert all(isinstance(k, int) for k in translation), "Expected integer keys"
        t_from = torch.tensor(translation.keys(), dtype=torch.int64, device=x.device)
        t_to = torch.tensor(translation.values(), device=x.device)
    elif isinstance(translation, torch.Tensor):
        assert translation.ndim == 2, (
            f"Expected 2D tensor, got {translation.ndim}D tensor"
        )
        t_from = translation[0, :]
        t_to = translation[1, :]
    else:
        assert isinstance(translation, tuple) and len(translation) == 2, (
            "Expected tuple of length 2"
        )
        t_from, t_to = translation
        assert t_from.ndim == 1 and t_to.ndim == 1, "Expected 1D tensors"

    # Ensure that values in `x` are in `t_from`
    if default is not None:
        v_unique = torch.unique(x)
        t_missing = ~torch.isin(v_unique, t_from)

        # Add missing values
        t_from = torch.cat([t_from, v_unique[t_missing]])
        t_to = torch.cat([t_to, torch.full_like(v_unique[t_missing], default)])

        # Sort the values (for bucketize)
        t_from_sort_indices = torch.argsort(t_from)
        t_from = t_from[t_from_sort_indices]
        t_to = t_to[t_from_sort_indices]

    assert t_from.shape == t_to.shape, "Expected tensors of the same shape"
    assert torch.all(torch.isin(x, t_from)), (
        f"Not all values in `x` ({x.detach().unique().cpu().tolist()}) are in `t_from` ({t_from.detach().cpu().tolist()}))"
    )

    # Map the values
    i = torch.bucketize(x.ravel(), t_from)
    y = t_to[i]

    return y.reshape(x.shape)


def nanvar(t: Tensor, dim: int, keepdim=False):
    return (
        (t - t.nanmean(dim=dim, keepdim=True))
        .abs()
        .pow(2)
        .nanmean(dim=dim, keepdim=keepdim)
    )


def nanstd(t: Tensor, dim: int, keepdim=False):
    return torch.sqrt(nanvar(t, dim, keepdim))



from __future__ import annotations

import math

import torch

DEFAULT_ACOS_BOUND: float = 1.0 - 1e-4


def acos_linear_extrapolation(
    x: torch.Tensor,
    bounds: tuple[float, float] = (-DEFAULT_ACOS_BOUND, DEFAULT_ACOS_BOUND),
) -> torch.Tensor:
    """
    Implements `arccos(x)` which is linearly extrapolated outside `x`'s original
    domain of `(-1, 1)`. This allows for stable backpropagation in case `x`
    is not guaranteed to be strictly within `(-1, 1)`.

    More specifically:
    ```
    bounds=(lower_bound, upper_bound)
    if lower_bound <= x <= upper_bound:
        acos_linear_extrapolation(x) = acos(x)
    elif x <= lower_bound: # 1st order Taylor approximation
        acos_linear_extrapolation(x)
            = acos(lower_bound) + dacos/dx(lower_bound) * (x - lower_bound)
    else:  # x >= upper_bound
        acos_linear_extrapolation(x)
            = acos(upper_bound) + dacos/dx(upper_bound) * (x - upper_bound)
    ```

    Args:
        x: Input `Tensor`.
        bounds: A float 2-tuple defining the region for the
            linear extrapolation of `acos`.
            The first/second element of `bound`
            describes the lower/upper bound that defines the lower/upper
            extrapolation region, i.e. the region where
            `x <= bound[0]`/`bound[1] <= x`.
            Note that all elements of `bound` have to be within (-1, 1).

    Returns
    -------
        acos_linear_extrapolation: `Tensor` containing the extrapolated `arccos(x)`.
    """
    lower_bound, upper_bound = bounds

    if lower_bound > upper_bound:
        msg = "lower bound has to be smaller or equal to upper bound."
        raise ValueError(msg)

    if lower_bound <= -1.0 or upper_bound >= 1.0:
        msg = "Both lower bound and upper bound have to be within (-1, 1)."
        raise ValueError(msg)

    # init an empty tensor and define the domain sets
    acos_extrap = torch.empty_like(x)
    x_upper = x >= upper_bound
    x_lower = x <= lower_bound
    x_mid = (~x_upper) & (~x_lower)

    # acos calculation for upper_bound < x < lower_bound
    acos_extrap[x_mid] = torch.acos(x[x_mid])
    # the linear extrapolation for x >= upper_bound
    acos_extrap[x_upper] = _acos_linear_approximation(x[x_upper], upper_bound)
    # the linear extrapolation for x <= lower_bound
    acos_extrap[x_lower] = _acos_linear_approximation(x[x_lower], lower_bound)

    return acos_extrap


def _acos_linear_approximation(x: torch.Tensor, x0: float) -> torch.Tensor:
    """Calculates the 1st order Taylor expansion of `arccos(x)` around `x0`."""
    return (x - x0) * _dacos_dx(x0) + math.acos(x0)


def _dacos_dx(x: float) -> float:
    """Calculates the derivative of `arccos(x)` w.r.t. `x`."""
    return (-1.0) / math.sqrt(1.0 - x * x)



from __future__ import annotations

import torch
from torch import Tensor


def get_gaussian_2d(
    kernel_size: int | tuple[int, int], sigma: float | tuple[float, float]
) -> Tensor:
    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)
    if isinstance(sigma, float | int):
        sigma = (float(sigma), float(sigma))

    kx, ky = kernel_size
    sx, sy = sigma

    x = torch.arange(kx).float() - (kx - 1) / 2
    y = torch.arange(ky).float() - (ky - 1) / 2

    gauss_x = torch.exp(-(x**2) / (2 * sx**2))
    gauss_y = torch.exp(-(y**2) / (2 * sy**2))

    gauss_2d = gauss_x.view(1, -1) * gauss_y.view(-1, 1)
    gauss_2d = gauss_2d / gauss_2d.sum()

    return gauss_2d



"""
Numerical stability.
-------------------
Some implementations use `torch.finfo(torch.float32).eps` to determine the epsilon
value used to enforce numeric stability. This value is the smallest positive
floating point value that can be represented in the given floating point type.
However, this value is not always the best choice for enforcing numeric stability
in all cases. For example, when comparing two floating point numbers, the difference
between the two numbers should be greater than the epsilon value to be considered
significant. In this case, a value of `1e-6` is used as the default epsilon value
to enforce numeric stability (i.e. the resolution of the floating point numbers).

For exponential functions, the input value is clamped to a maximum
of `88.0` to prevent overflow. This value is chosen because `exp(88.7)` is the
largest value that can be represented in 32-bit floating point precision without
overflowing.

Generally, we follow the following rules for numeric stability:
    - Division by zero is prevented by **adding** a small epsilon value to the denominator.
    - Logarithm of zero is prevented by **adding** a small epsilon value to the input.
    - Square root of negative values is prevented by **clamping** the input to a small value.
    - Exponential functions are **clamped **to a maximum input value to prevent overflow.
"""

import math
from typing import Final

import torch

from unipercept.types import Tensor

__all__ = []  # all symbols are not exported

EPS_FP32: Final = torch.finfo(torch.float32).eps
EPS_DIV_FP32: Final = torch.finfo(torch.float32).eps
MAX_EXP_FP32: Final = math.floor(math.log(torch.finfo(torch.float32).max))
EPS_FP16: Final = torch.finfo(torch.float16).eps
MAX_EXP_FP16: Final = math.floor(math.log(torch.finfo(torch.float16).max))
EPS_BF16: Final = torch.finfo(torch.bfloat16).eps
MAX_EXP_BF16: Final = math.floor(math.log(torch.finfo(torch.bfloat16).max))


def div(a: Tensor, b: Tensor, *, eps: float = EPS_DIV_FP32) -> Tensor:
    return a / (b + eps)


def log(x: Tensor, *, eps: float) -> Tensor:
    return torch.log(x + eps)


def sqrt(x: Tensor, *, eps: float) -> Tensor:
    return x.clamp_min(eps).sqrt()


def exp(x: Tensor, *, max_exp: float) -> Tensor:
    return x.clamp_max(max_exp).exp()



"""
Register tensors belonging to certain categories.

The following categories are available:

- `pixel_maps`: Pixel maps are tensors that represent 2D (image/pixel) data
- `point_maps`: Point maps are tensors that represent 3D point cloud data
"""

from collections.abc import Callable
from typing import TypeVar

import torch

T = TypeVar("T")


class SimpleRegistry[T]:
    def __init__(self, check: Callable[[type[T]], bool] | None = None):
        self._items: set[type[T]] = set()
        self._check = check

    def add(self, item: type[T]):
        if self._check and not self._check(item):
            raise TypeError(f"Item {item} failed registry check.")
        self._items.add(item)

    def register(self):
        def decorator(item: type[T]):
            self.add(item)
            return item

        return decorator

    def __iter__(self):
        return iter(self._items)

    def __contains__(self, item):
        return item in self._items


__all__ = ["pixel_maps", "point_maps"]


def _check_tensorsubclass(cls: type) -> bool:
    return isinstance(cls, type) and issubclass(cls, torch.Tensor)


pixel_maps = SimpleRegistry[torch.Tensor](check=_check_tensorsubclass)
point_maps = SimpleRegistry[torch.Tensor](check=_check_tensorsubclass)



"""Various utilities for working with geometry transformations and conversions.

The API is rougly intended to match the conventions proposed
in `PyTorch3D <https://pytorch3d.readthedocs.io/>`_.

Further reading:
- `Quaternions <https://danceswithcode.net/engineeringnotes/quaternions/quaternions.html>`_

Notes
-----
Parts of this module are generated via `GitHub Copilot<https://copilot.github.com/>`_
and require manual review and testing. Contributions are welcome.
"""

import enum as E
import typing

import torch
import torch.fx
from torch import nn

from unipercept.types import Device, DType, Size, Tensor


class AxesConvention(E.StrEnum):
    """Enum for conventions on how to orient the world coordinate system.

    The following are implemented:

        - ``OPENCV`` : ``(+x, +y, +z) = (right   , down  , fwd    )`` (right-handed)
        - ``ISO8855``: ``(+x, +y, +z) = (fwd     , left  , up     )`` (right-handed)
        - ``OPENGL`` : ``(+x, +y, +z) = (right   , up    , bwd    )`` (right-handed)
        - ``OPEN3D`` : ``(+x, +y, +z) = (right   , up    , bwd    )`` (right-handed)
        - ``DIRECTX``: ``(+x, +y, +z) = (right   , up    , fwd    )`` (left-handed)

    Notes
    -----
    The camera coordinate system is always oriented as ``OPENCV``, and pixels are
    oriented as ``(+u, +v) = (right, down)`` in the image plane, with (0, 0) at the
    top-left corner.

    See Also
    --------
    - `OpenCV documentation <https://docs.opencv.org/3.4/d9/d0c/group__calib3d.html>`_
    - `Cityscapes calibration <https://github.com/mcordts/cityscapesScripts>`_
    - `Learn OpenGL <https://learnopengl.com/Getting-started/Coordinate-Systems>`_
    """

    ISO8855 = E.auto()
    OPENCV = E.auto()
    OPENGL = E.auto()
    OPEN3D = E.auto()
    DIRECTX = E.auto()


def _transform_to_opencv(
    convention: AxesConvention | str,
    device: Device | None = None,
    dtype: DType = torch.float32,
) -> Tensor:
    match convention:
        case AxesConvention.OPENCV:
            # (X, Y, Z) = [right, down, fwd]
            return torch.eye(4, device=device, dtype=dtype)
        case AxesConvention.ISO8855:
            # (X, Y, Z) = [fwd, left, up]
            return torch.tensor(
                [
                    [0.0, 0.0, 1.0, 0.0],  # X -> Z
                    [-1.0, 0.0, 0.0, 0.0],  # Y -> -X
                    [0.0, -1.0, 0.0, 0.0],  # Z -> -Y
                    [0.0, 0.0, 0.0, 1.0],  # 1 -> 1 (homogeneous)
                ],
                device=device,
                dtype=dtype,
            )
        case AxesConvention.OPENGL | AxesConvention.OPEN3D:
            # (X, Y, Z) = [right, up, bw]
            return torch.tensor(
                [
                    [1.0, 0.0, 0.0, 0.0],  # X -> X
                    [0.0, -1.0, 0.0, 0.0],  # Y -> -Y
                    [0.0, 0.0, -1.0, 0.0],  # Z -> -Z
                    [0.0, 0.0, 0.0, 1.0],  # 1 -> 1 (homogeneous)
                ],
                device=device,
                dtype=dtype,
            )
        case _:
            msg = f"Unknown convention {convention!r}!"
            raise ValueError(msg)


def _mapping_to_opencv(
    convention: AxesConvention | str,
    device: Device | None = None,
    dtype: DType = torch.float32,
) -> Tensor:
    return _transform_to_opencv(convention, device=device, dtype=dtype)[:3, :3]


def convert_extrinsics(
    extrinsics: Tensor,
    *,
    src: AxesConvention | str = AxesConvention.OPENCV,
    tgt: AxesConvention | str = AxesConvention.OPENCV,
) -> Tensor:
    """
    Changes an extrinsic transformation to a different world axes convention.

    This is useful in cases where a camera/dataset provides the extrinsic parameters
    in a different world convention than ours. For example, the Cityscapes dataset
    provides their extrinsics in ISO8855 contention, while we use OpenCV. This function
    allows using the direct parameters from the dataset to build the extrinsics matrix,
    which can subsequently be converted to OpenCV for use in our system.

    Parameters
    ----------
    extrinsics: Tensor[*, 4, 4]
        Input matrix.
    src: AxesConvention or str
        Current convention. See :class:`AxesConvention` for possible values.
    tgt: AxesConvention or str
        Target convention. See :class:`AxesConvention` for possible values.

    Returns
    -------
    Tensor[*, 4, 4]
        Output matrix.

    Examples
    --------
    >>> extrinsics = torch.eye(4)
    >>> converted_extrinsics = convert_extrinsics(
    ...     extrinsics, src=AxesConvention.ISO8855, tgt=AxesConvention.OPENCV
    ... )
    >>> print(converted_extrinsics)
    tensor([[ 0.,  0.,  1.,  0.],
            [-1.,  0.,  0.,  0.],
            [ 0., -1.,  0.,  0.],
            [ 0.,  0.,  0.,  1.]])
    """
    if src == tgt:
        return extrinsics

    device = extrinsics.device
    dtype = extrinsics.dtype
    T_src = _transform_to_opencv(src, device=device, dtype=dtype).T
    T_tgt = _transform_to_opencv(tgt, device=device, dtype=dtype)

    return extrinsics @ T_src @ T_tgt


def convert_points(
    points: torch.Tensor,
    *,
    src: AxesConvention | str = AxesConvention.OPENCV,
    tgt: AxesConvention | str = AxesConvention.OPENCV,
) -> torch.Tensor:
    """
    Converts points from one world axes convention to another.

    This is useful in cases where we want to display points in a different world
    convention than the one they were projected into.
    For example, if we have a projection onto OpenCV coordinates (the default) and we
    want to display a point cloud in Matplotlib (which uses Z-up coordinates), then
    we can convert the projected points to ISO8855 to get the correct orientation.

    Parameters
    ----------
    points: Tensor[*, 3]
        Input points.
    src: AxesConvention or str
        Current convention. See :class:`AxesConvention` for possible values.
    tgt: AxesConvention or str
        Target convention. See :class:`AxesConvention` for possible values.

    Returns
    -------
    Tensor[*, 3]
        Output points.

    Examples
    --------
    >>> points = torch.tensor([[1.0, 2.0, 3.0]])
    >>> converted_points = convert_points(
    ...     points, src=AxesConvention.ISO8855, tgt=AxesConvention.OPENCV
    ... )
    >>> print(converted_points)
    tensor([[ 3., -1., -2.]])
    """
    if src == tgt:
        return points

    device = points.device
    dtype = points.dtype
    R_src = _mapping_to_opencv(src, device=device, dtype=dtype)
    R_tgt = _mapping_to_opencv(tgt, device=device, dtype=dtype).T
    return points @ R_src @ R_tgt


def extrinsics_from_parameters(
    angles: list[tuple[float, float, float]] | Tensor,
    translation: list[tuple[float, float, float]] | Tensor,
    convention: AxesConvention | str = AxesConvention.ISO8855,
) -> Tensor:
    """
    Build the extrinsic matrix (R|t) using an angles vector and translations.

    Parameters
    ----------
    angles: list[tuple[float, float, float]] or Tensor
        List of tuples or tensor containing the rotation angles.
    translation: list[tuple[float, float, float]] or Tensor
        List of tuples or tensor containing the translation vectors.
    convention: AxesConvention or str
        Convention to use for the extrinsic matrix. See :class:`AxesConvention` for possible values.

    Returns
    -------
    Tensor[*, 4, 4]
        Extrinsic matrix.

    Examples
    --------
    >>> angles = [(0.0, 0.0, 0.0)]
    >>> translation = [(0.0, 0.0, 0.0)]
    >>> extrinsics = extrinsics_from_parameters(
    ...     angles, translation, convention=AxesConvention.ISO8855
    ... )
    >>> print(extrinsics)
    tensor([[ 0.,  0.,  1.,  0.],
            [-1.,  0.,  0.,  0.],
            [ 0., -1.,  0.,  0.],
            [ 0.,  0.,  0.,  1.]])
    """
    rotation = torch.as_tensor(angles).float()
    translation = torch.as_tensor(translation).float()

    # Amount of cameras
    ndim = rotation.ndim
    assert ndim == translation.ndim
    if ndim == 1:
        rotation = rotation.unsqueeze(0)
        translation = translation.unsqueeze(0)

    N = rotation.size(0)

    # Value checks
    assert rotation.shape[-1] == 3, rotation.shape
    assert translation.shape[-1] == 3, translation.shape

    # Create extrinsic matrix [R|t] + [0 0 0 1]
    extrinsic_matrix = torch.zeros(N, 4, 4, device=rotation[0].device)
    extrinsic_matrix[:, :3, :3] = axis_angle_to_rotation(rotation)
    for i in range(N):
        extrinsic_matrix[i, :3, 3] = translation[i]
        extrinsic_matrix[i, 3, 3] = 1.0

    if ndim == 1:
        extrinsic_matrix = extrinsic_matrix.squeeze(0)

    return convert_extrinsics(
        extrinsic_matrix, src=convention, tgt=AxesConvention.OPENCV
    )


def intrinsics_from_parameters(
    focal_length: list[tuple[float, float]] | Tensor,
    principal_point: list[tuple[float, float]] | Tensor,
    orthographic: bool,
) -> Tensor:
    """
    Build the calibration matrix (K-matrix) using the focal lengths and
    principal points.

    Parameters
    ----------
    focal_length: list[tuple[float, float]] or Tensor
        List of tuples or tensor containing the focal lengths.
    principal_point: list[tuple[float, float]] or Tensor
        List of tuples or tensor containing the principal points.
    orthographic: bool
        Whether to use orthographic projection.

    Returns
    -------
    Tensor[*, 4, 4]
        Calibration matrix.

    Examples
    --------
    >>> focal_length = [(1000.0, 1000.0)]
    >>> principal_point = [(500.0, 500.0)]
    >>> K = intrinsics_from_parameters(
    ...     focal_length, principal_point, orthographic=False
    ... )
    >>> print(K)
    tensor([[1000.,    0.,  500.,    0.],
            [   0., 1000.,  500.,    0.],
            [   0.,    0.,    1.,    1.],
            [   0.,    0.,    0.,    1.]])
    """
    # Focal length, shape (), (N, 1) or (N, 2)
    if not isinstance(focal_length, torch.Tensor):
        focal_length = torch.as_tensor(focal_length).float()
    if focal_length.shape[-1] == 1:
        fx = fy = focal_length
    else:
        fx, fy = focal_length.unbind(-1)

    # Principal point
    if not isinstance(principal_point, torch.Tensor):
        principal_point = torch.as_tensor(principal_point).float()
    px, py = principal_point.unbind(-1)

    # Batched/unbatched
    ndim = focal_length.ndim
    assert ndim == principal_point.ndim
    if ndim == 1:
        fx = fx.unsqueeze(0)
        fy = fy.unsqueeze(0)
        px = px.unsqueeze(0)
        py = py.unsqueeze(0)

    # Amount of cameras
    N = fx.size(0)

    # Create calibration matrix
    K: Tensor = fx.new_zeros(N, 4, 4)
    K[:, 0, 0] = fx
    K[:, 1, 1] = fy
    if orthographic:
        K[:, 0, 3] = px
        K[:, 1, 3] = py
        K[:, 2, 2] = 1.0
        K[:, 3, 3] = 1.0
    else:
        K[:, 0, 2] = px
        K[:, 1, 2] = py
        K[:, 3, 2] = 1.0
        K[:, 2, 3] = 1.0

    if ndim == 1:
        K = K.squeeze(0)

    return K


# @torch.fx.wrap
@torch.autocast("cuda", enabled=False)
def unsafe_inverse(x: Tensor) -> Tensor:
    """
    Compute the inverse of the input tensor. The input is cast to FP32 and back
    to the input datatype if it is not supported by the inversion function.

    Parameters
    ----------
    x: Tensor
        Input tensor.

    Returns
    -------
    Tensor
        Inverse of the input tensor.

    Examples
    --------
    >>> x = torch.eye(3)
    >>> x_inv = unsafe_inverse(x)
    >>> print(x_inv)
    tensor([[1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]])
    """
    x_inv = typing.cast(Tensor, torch.linalg.inv_ex(x.float()).inverse)  # type: ignore[attr-defined]
    return x_inv.type_as(x)


def rad2deg(tensor: Tensor) -> Tensor:
    """
    Function that converts angles from radians to degrees.

    Parameters
    ----------
    tensor: Tensor
        Tensor of arbitrary shape.

    Returns
    -------
    Tensor
        Tensor with same shape as input.

    Examples
    --------
    >>> input = torch.tensor(3.1415926535)
    >>> rad2deg(input)
    tensor(180.)
    """
    return 180.0 * tensor / torch.pi


def deg2rad(tensor: Tensor) -> Tensor:
    """
    Function that converts angles from degrees to radians.

    Parameters
    ----------
    tensor: Tensor
        Tensor of arbitrary shape.

    Returns
    -------
    Tensor
        Tensor with same shape as input.

    Examples
    --------
    >>> input = torch.tensor(180.0)
    >>> deg2rad(input)
    tensor(3.1416)
    """
    return tensor * torch.pi / 180.0


def pol2cart(rho: Tensor, phi: Tensor) -> tuple[Tensor, Tensor]:
    """
    Function that converts polar coordinates to cartesian coordinates.

    Parameters
    ----------
    rho: Tensor
        Tensor of arbitrary shape.
    phi: Tensor
        Tensor of same arbitrary shape.

    Returns
    -------
    tuple[Tensor, Tensor]
        - x: Tensor with same shape as input.
        - y: Tensor with same shape as input.

    Examples
    --------
    >>> rho = torch.rand(1, 3, 3)
    >>> phi = torch.rand(1, 3, 3)
    >>> x, y = pol2cart(rho, phi)
    """
    x = rho * phi.cos()
    y = rho * phi.sin()
    return x, y


def cart2pol(x: Tensor, y: Tensor, eps: float = 1e-8) -> tuple[Tensor, Tensor]:
    """
    Function that converts cartesian coordinates to polar coordinates.

    Parameters
    ----------
    x: Tensor
        Tensor of arbitrary shape.
    y: Tensor
        Tensor of same arbitrary shape.
    eps: float
        To avoid division by zero.

    Returns
    -------
    tuple[Tensor, Tensor]
        - rho: Tensor with same shape as input.
        - phi: Tensor with same shape as input.

    Examples
    --------
    >>> x = torch.rand(1, 3, 3)
    >>> y = torch.rand(1, 3, 3)
    >>> rho, phi = cart2pol(x, y)
    """
    rho = torch.sqrt(x**2 + y**2 + eps)
    phi = torch.atan2(y, x)
    return rho, phi


def euclidean_to_homogeneous_points(points: Tensor) -> Tensor:
    """
    Convert Euclidean points to homogeneous coordinates.

    Usually this is done for a set of 2D points before reprojecting them to 3D points.
    This simply entails adding a column of ones to the input points.

    Parameters
    ----------
    points: Tensor[*, N, D]
        Input points to homogenize.

    Returns
    -------
    Tensor[*, N, D+1]
        Homogeneous coordinates for each point.

    Examples
    --------
    >>> points = torch.tensor([[1.0, 2.0], [3.0, 4.0]])
    >>> homogeneous_points = euclidean_to_homogeneous_points(points)
    >>> print(homogeneous_points)
    tensor([[1., 2., 1.],
            [3., 4., 1.]])
    """
    return nn.functional.pad(points, (0, 1), "constant", value=1.0)


def homogeneous_to_euclidean_points(points: Tensor, eps: float = 1e-8) -> Tensor:
    """Convert homogeneous points to Euclidean space.

    Semantically, this inverts the result of :func:`_euclidean_to_homogeneous_points`.

    Parameters
    ----------
    points: Tensor[*, N, D+1]
        Points to convert.
    eps: float, optional
        Small value to avoid division by zero.

    Returns
    -------
    Tensor[*, N, D]
        Euclidean coordinates for each point.

    Examples
    --------
    >>> points = torch.tensor([[1.0, 2.0, 1.0], [3.0, 4.0, 1.0]])
    >>> euclidean_points = homogeneous_to_euclidean_points(points)
    >>> print(euclidean_points)
    tensor([[1., 2.],
            [3., 4.]])
    """
    z_vec = points[..., -1:]
    mask = torch.abs(z_vec) > eps
    scale = torch.where(mask, 1.0 / (z_vec + eps), 1.0)

    return scale * points[..., :-1]


def convert_affinematrix_to_homography(A: Tensor) -> Tensor:
    """
    Function that converts batch of affine matrices.

    Parameters
    ----------
    A: Tensor[*, 2, 3]
        Affine matrix.

    Returns
    -------
    Tensor[*, 3, 3]
        Homography matrix.

    Examples
    --------
    >>> A = torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
    >>> H = convert_affinematrix_to_homography(A)
    >>> print(H)
    tensor([[1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]])
    """
    H: Tensor = nn.functional.pad(A, [0, 0, 0, 1], "constant", value=0.0)
    H[..., -1, -1] += 1.0
    return H


def axis_angle_to_rotation(axis_angle: Tensor) -> Tensor:
    """
    Convert 3d vector of axis-angle rotation to 3x3 rotation matrix.

    Parameters
    ----------
    axis_angle: Tensor[*, 3]
        Axis-angle rotations in radians.

    Returns
    -------
    Tensor[*, 3, 3]
        Rotation matrix

    Examples
    --------
    >>> input = torch.tensor([[0.0, 0.0, 0.0]])
    >>> axis_angle_to_rotation(input)
    tensor([[[1., 0., 0.],
             [0., 1., 0.],
             [0., 0., 1.]]])

    >>> input = torch.tensor([[1.5708, 0.0, 0.0]])
    >>> axis_angle_to_rotation(input)
    tensor([[[ 1.0000e+00,  0.0000e+00,  0.0000e+00],
             [ 0.0000e+00, -3.6200e-06, -1.0000e+00],
             [ 0.0000e+00,  1.0000e+00, -3.6200e-06]]])
    """

    def _compute_rotation(
        axis_angle: Tensor, theta2: Tensor, eps: float = 1e-6
    ) -> Tensor:
        k_one = 1.0
        theta = torch.sqrt(theta2)
        wxyz = axis_angle / (theta + eps)
        wx, wy, wz = torch.chunk(wxyz, 3, dim=1)
        cos_theta = torch.cos(theta)
        sin_theta = torch.sin(theta)

        r00 = cos_theta + wx * wx * (k_one - cos_theta)
        r10 = wz * sin_theta + wx * wy * (k_one - cos_theta)
        r20 = -wy * sin_theta + wx * wz * (k_one - cos_theta)
        r01 = wx * wy * (k_one - cos_theta) - wz * sin_theta
        r11 = cos_theta + wy * wy * (k_one - cos_theta)
        r21 = wx * sin_theta + wy * wz * (k_one - cos_theta)
        r02 = wy * sin_theta + wx * wz * (k_one - cos_theta)
        r12 = -wx * sin_theta + wy * wz * (k_one - cos_theta)
        r22 = cos_theta + wz * wz * (k_one - cos_theta)
        rotation = torch.cat([r00, r01, r02, r10, r11, r12, r20, r21, r22], dim=1)
        return rotation.view(-1, 3, 3)

    def _compute_rotation_taylor(axis_angle: Tensor) -> Tensor:
        rx, ry, rz = torch.chunk(axis_angle, 3, dim=1)
        k_one = torch.ones_like(rx)
        rotation = torch.cat([k_one, -rz, ry, rz, k_one, -rx, -ry, rx, k_one], dim=1)
        return rotation.view(-1, 3, 3)

    _axis_angle = torch.unsqueeze(axis_angle, dim=1)
    theta2 = torch.matmul(_axis_angle, _axis_angle.transpose(1, 2))
    theta2 = torch.squeeze(theta2, dim=1)

    # Compute rotation matrices
    rotation_normal = _compute_rotation(axis_angle, theta2)
    rotation_taylor = _compute_rotation_taylor(axis_angle)

    # Create mask to handle both cases
    eps = 1e-6
    mask = (theta2 > eps).view(-1, 1, 1).to(theta2.device)
    mask_pos = (mask).type_as(theta2)
    mask_neg = (~mask).type_as(theta2)

    # Create output pose matrix with masked values
    return mask_pos * rotation_normal + mask_neg * rotation_taylor


def rotation_to_axis_angle(rotation: Tensor) -> Tensor:
    """
    Convert 3x3 rotation matrix to Rodrigues vector in radians.

    Parameters
    ----------
    rotation: Tensor[*, 3, 3]
        Rotation matrix.

    Returns
    -------
    Tensor[*, 3]
        Rodrigues vector transformation.

    Examples
    --------
    >>> input = torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    >>> rotation_to_axis_angle(input)
    tensor([0., 0., 0.])

    >>> input = torch.tensor([[1.0, 0.0, 0.0], [0.0, 0.0, -1.0], [0.0, 1.0, 0.0]])
    >>> rotation_to_axis_angle(input)
    tensor([1.5708, 0.0000, 0.0000])
    """
    quaternion: Tensor = rotation_to_quaternion(rotation)
    return quaternion_to_axis_angle(quaternion)


def rotation_to_quaternion(rotation: Tensor, eps: float = 1e-8) -> Tensor:
    """
    Convert 3x3 rotation matrix to 4d quaternion vector.

    The quaternion vector has components in (w, x, y, z) format.

    Parameters
    ----------
    rotation: Tensor[*, 3, 3]
        The rotation matrix to convert.
    eps: float
        Small value to avoid zero division.

    Returns
    -------
    Tensor[*, 4]
        The rotation in quaternion.

    Examples
    --------
    >>> rotation = torch.tensor([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
    >>> quaternion = rotation_to_quaternion(rotation)
    >>> print(quaternion)
    tensor([1., 0., 0., 0.])
    """

    def safe_zero_division(numerator: Tensor, denominator: Tensor) -> Tensor:
        eps: float = torch.finfo(numerator.dtype).tiny
        return numerator / torch.clamp(denominator, min=eps)

    rotation_vec: Tensor = rotation.reshape(*rotation.shape[:-2], 9)

    m00, m01, m02, m10, m11, m12, m20, m21, m22 = torch.chunk(
        rotation_vec, chunks=9, dim=-1
    )

    trace: Tensor = m00 + m11 + m22

    def trace_positive_cond() -> Tensor:
        sq = torch.sqrt(trace + 1.0 + eps) * 2.0  # sq = 4 * qw.
        qw = 0.25 * sq
        qx = safe_zero_division(m21 - m12, sq)
        qy = safe_zero_division(m02 - m20, sq)
        qz = safe_zero_division(m10 - m01, sq)
        return torch.cat((qw, qx, qy, qz), dim=-1)

    def cond_1() -> Tensor:
        sq = torch.sqrt(1.0 + m00 - m11 - m22 + eps) * 2.0  # sq = 4 * qx.
        qw = safe_zero_division(m21 - m12, sq)
        qx = 0.25 * sq
        qy = safe_zero_division(m01 + m10, sq)
        qz = safe_zero_division(m02 + m20, sq)
        return torch.cat((qw, qx, qy, qz), dim=-1)

    def cond_2() -> Tensor:
        sq = torch.sqrt(1.0 + m11 - m00 - m22 + eps) * 2.0  # sq = 4 * qy.
        qw = safe_zero_division(m02 - m20, sq)
        qx = safe_zero_division(m01 + m10, sq)
        qy = 0.25 * sq
        qz = safe_zero_division(m12 + m21, sq)
        return torch.cat((qw, qx, qy, qz), dim=-1)

    def cond_3() -> Tensor:
        sq = torch.sqrt(1.0 + m22 - m00 - m11 + eps) * 2.0  # sq = 4 * qz.
        qw = safe_zero_division(m10 - m01, sq)
        qx = safe_zero_division(m02 + m20, sq)
        qy = safe_zero_division(m12 + m21, sq)
        qz = 0.25 * sq
        return torch.cat((qw, qx, qy, qz), dim=-1)

    where_2 = torch.where(m11 > m22, cond_2(), cond_3())
    where_1 = torch.where((m00 > m11) & (m00 > m22), cond_1(), where_2)

    quaternion: Tensor = torch.where(trace > 0.0, trace_positive_cond(), where_1)
    return quaternion


def normalize_quaternion(quaternion: Tensor, eps: float = 1e-12) -> Tensor:
    """
    Normalize a quaternion.

    The quaternion should be in (x, y, z, w) or (w, x, y, z) format.

    Parameters
    ----------
    quaternion: Tensor
        A tensor containing a quaternion to be normalized.
        The tensor can be of shape :math:`(*, 4)`.
    eps: float
        Small value to avoid division by zero.

    Returns
    -------
    Tensor
        The normalized quaternion of shape :math:`(*, 4)`.

    Examples
    --------
    >>> quaternion = torch.tensor((1.0, 0.0, 1.0, 0.0))
    >>> normalize_quaternion(quaternion)
    tensor([0.7071, 0.0000, 0.7071, 0.0000])
    """
    return nn.functional.normalize(quaternion, p=2.0, dim=-1, eps=eps)


def quaternion_to_rotation(quaternion: Tensor) -> Tensor:
    """
    Convert a quaternion to a rotation matrix.

    The quaternion should be in (w, x, y, z) format.

    Parameters
    ----------
    quaternion: Tensor
        A tensor containing a quaternion to be converted.
        The tensor can be of shape :math:`(*, 4)`.

    Returns
    -------
    Tensor[*, 3, 3]
        The rotation matrix.

    Examples
    --------
    >>> quaternion = torch.tensor([1.0, 0.0, 0.0, 0.0])
    >>> rotation = quaternion_to_rotation(quaternion)
    >>> print(rotation)
    tensor([[1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]])
    """
    if not isinstance(quaternion, torch.Tensor):
        msg = f"Input type is not a Tensor. Got {type(quaternion)}"
        raise TypeError(msg)

    if quaternion.shape[-1] != 4:
        msg = f"Input must be a tensor of shape (*, 4). Got {quaternion.shape}"
        raise ValueError(msg)

    # normalize the input quaternion
    quaternion_norm: Tensor = normalize_quaternion(quaternion)

    # unpack the normalized quaternion components
    w = quaternion_norm[..., 0]
    x = quaternion_norm[..., 1]
    y = quaternion_norm[..., 2]
    z = quaternion_norm[..., 3]

    # compute the actual conversion
    tx: Tensor = 2.0 * x
    ty: Tensor = 2.0 * y
    tz: Tensor = 2.0 * z
    twx: Tensor = tx * w
    twy: Tensor = ty * w
    twz: Tensor = tz * w
    txx: Tensor = tx * x
    txy: Tensor = ty * x
    txz: Tensor = tz * x
    tyy: Tensor = ty * y
    tyz: Tensor = tz * y
    tzz: Tensor = tz * z
    one: Tensor = torch.tensor(1.0, device=quaternion.device, dtype=quaternion.dtype)

    matrix_flat: Tensor = torch.stack(
        (
            one - (tyy + tzz),
            txy - twz,
            txz + twy,
            txy + twz,
            one - (txx + tzz),
            tyz - twx,
            txz - twy,
            tyz + twx,
            one - (txx + tyy),
        ),
        dim=-1,
    )

    # this slightly awkward construction of the output shape is to satisfy torchscript
    output_shape = [*list(quaternion.shape[:-1]), 3, 3]
    return matrix_flat.reshape(output_shape)


def quaternion_to_axis_angle(quaternion: Tensor) -> Tensor:
    """
    Convert quaternion vector to axis angle of rotation in radians.

    The quaternion should be in (w, x, y, z) format.

    Parameters
    ----------
    quaternion: Tensor
        Tensor with quaternions.

    Returns
    -------
    Tensor
        Tensor with axis angle of rotation.

    Examples
    --------
    >>> quaternion = torch.tensor([1.0, 0.0, 0.0, 0.0])
    >>> axis_angle = quaternion_to_axis_angle(quaternion)
    >>> print(axis_angle)
    tensor([0., 0., 0.])
    """
    if not torch.is_tensor(quaternion):
        msg = f"Input type is not a Tensor. Got {type(quaternion)}"
        raise TypeError(msg)

    if quaternion.shape[-1] != 4:
        msg = f"Input must be a tensor of shape Nx4 or 4. Got {quaternion.shape}"
        raise ValueError(msg)

    cos_theta = quaternion[..., 0]
    q1 = quaternion[..., 1]
    q2 = quaternion[..., 2]
    q3 = quaternion[..., 3]

    sin_squared_theta: Tensor = q1 * q1 + q2 * q2 + q3 * q3

    sin_theta: Tensor = torch.sqrt(sin_squared_theta)
    two_theta: Tensor = 2.0 * torch.where(
        cos_theta < 0.0,
        torch.atan2(-sin_theta, -cos_theta),
        torch.atan2(sin_theta, cos_theta),
    )

    k_pos: Tensor = two_theta / sin_theta
    k_neg: Tensor = 2.0 * torch.ones_like(sin_theta)
    k: Tensor = torch.where(sin_squared_theta > 0.0, k_pos, k_neg)

    axis_angle: Tensor = torch.zeros_like(quaternion)[..., :3]
    axis_angle[..., 0] += q1 * k
    axis_angle[..., 1] += q2 * k
    axis_angle[..., 2] += q3 * k
    return axis_angle


def quaternion_log_to_exp(quaternion: Tensor, eps: float = 1e-8) -> Tensor:
    """
    Apply exponential map to log quaternion.

    The quaternion should be in (w, x, y, z) format.

    Parameters
    ----------
    quaternion: Tensor
        A tensor containing a quaternion to be converted.
        The tensor can be of shape :math:`(*, 3)`.
    eps: float
        A small number for clamping.

    Returns
    -------
    Tensor
        The quaternion exponential map of shape :math:`(*, 4)`.

    Examples
    --------
    >>> quaternion = torch.tensor([0.0, 1.0, 0.0])
    >>> quaternion_exp = quaternion_log_to_exp(quaternion)
    >>> print(quaternion_exp)
    tensor([0.5403, 0.0000, 0.8415, 0.0000])
    """
    # compute quaternion norm
    norm_q: Tensor = torch.norm(quaternion, p=2, dim=-1, keepdim=True).clamp(min=eps)

    # compute scalar and vector
    quaternion_vector: Tensor = quaternion * torch.sin(norm_q) / norm_q
    quaternion_scalar: Tensor = torch.cos(norm_q)

    # compose quaternion and return
    return torch.cat((quaternion_scalar, quaternion_vector), dim=-1)


def quaternion_exp_to_log(quaternion: Tensor, eps: float = 1e-8) -> Tensor:
    """
    Apply the log map to a quaternion.

    The quaternion should be in (w, x, y, z) format.

    Parameters
    ----------
    quaternion: Tensor
        A tensor containing a quaternion to be converted.
        The tensor can be of shape :math:`(*, 4)`.
    eps: float
        A small number for clamping.

    Returns
    -------
    Tensor
        The quaternion log map of shape :math:`(*, 3)`.

    Examples
    --------
    >>> quaternion = torch.tensor([0.5403, 0.0, 0.8415, 0.0])
    >>> quaternion_log = quaternion_exp_to_log(quaternion)
    >>> print(quaternion_log)
    tensor([0.0000, 1.0000, 0.0000])
    """
    # Unpack quaternion vector and scalar
    quaternion_scalar = quaternion[..., 0:1]
    quaternion_vector = quaternion[..., 1:4]

    # Compute quaternion norm
    norm_q: Tensor = torch.norm(quaternion_vector, p=2, dim=-1, keepdim=True).clamp(
        min=eps
    )

    # Apply log map
    quaternion_log: Tensor = (
        quaternion_vector
        * torch.acos(torch.clamp(quaternion_scalar, min=-1.0, max=1.0))
        / norm_q
    )

    return quaternion_log


def axis_angle_to_quaternion(axis_angle: Tensor) -> Tensor:
    """
    Convert an axis angle to a quaternion.

    The quaternion vector has components in (w, x, y, z) format.

    Adapted from ceres C++ library: ceres-solver/include/ceres/rotation.h

    Parameters
    ----------
    axis_angle: Tensor
        Tensor with axis angle in radians.

    Returns
    -------
    Tensor
        Tensor with quaternion.

    Examples
    --------
    >>> axis_angle = torch.tensor((0.0, 1.0, 0.0))
    >>> axis_angle_to_quaternion(axis_angle)
    tensor([0.8776, 0.0000, 0.4794, 0.0000])
    """
    if not torch.is_tensor(axis_angle):
        msg = f"Input type is not a Tensor. Got {type(axis_angle)}"
        raise TypeError(msg)

    if axis_angle.shape[-1] != 3:
        msg = f"Input must be a tensor of shape Nx3 or 3. Got {axis_angle.shape}"
        raise ValueError(msg)

    # unpack input and compute conversion
    a0: Tensor = axis_angle[..., 0:1]
    a1: Tensor = axis_angle[..., 1:2]
    a2: Tensor = axis_angle[..., 2:3]
    theta_squared: Tensor = a0 * a0 + a1 * a1 + a2 * a2

    theta: Tensor = torch.sqrt(theta_squared)
    half_theta: Tensor = theta * 0.5

    mask: Tensor = theta_squared > 0.0
    ones: Tensor = torch.ones_like(half_theta)

    k_neg: Tensor = 0.5 * ones
    k_pos: Tensor = torch.sin(half_theta) / theta
    k: Tensor = torch.where(mask, k_pos, k_neg)
    w: Tensor = torch.where(mask, torch.cos(half_theta), ones)

    quaternion: Tensor = torch.zeros(
        size=(*axis_angle.shape[:-1], 4),
        dtype=axis_angle.dtype,
        device=axis_angle.device,
    )
    quaternion[..., 1:2] = a0 * k
    quaternion[..., 2:3] = a1 * k
    quaternion[..., 3:4] = a2 * k
    quaternion[..., 0:1] = w
    return quaternion


def euler_from_quaternion(
    w: Tensor, x: Tensor, y: Tensor, z: Tensor
) -> tuple[Tensor, Tensor, Tensor]:
    """
    Convert a quaternion coefficients to Euler angles.

    Returned angles are in radians in XYZ convention.

    Parameters
    ----------
    w: Tensor
        Quaternion :math:`q_w` coefficient.
    x: Tensor
        Quaternion :math:`q_x` coefficient.
    y: Tensor
        Quaternion :math:`q_y` coefficient.
    z: Tensor
        Quaternion :math:`q_z` coefficient.

    Returns
    -------
    tuple[Tensor, Tensor, Tensor]
        A tuple with euler angles `roll`, `pitch`, `yaw`.

    Examples
    --------
    >>> w, x, y, z = (
    ...     torch.tensor(1.0),
    ...     torch.tensor(0.0),
    ...     torch.tensor(0.0),
    ...     torch.tensor(0.0),
    ... )
    >>> roll, pitch, yaw = euler_from_quaternion(w, x, y, z)
    >>> print(roll, pitch, yaw)
    tensor(0.) tensor(0.) tensor(0.)
    """
    yy = y * y

    sinr_cosp = 2.0 * (w * x + y * z)
    cosr_cosp = 1.0 - 2.0 * (x * x + yy)
    roll = sinr_cosp.atan2(cosr_cosp)

    sinp = 2.0 * (w * y - z * x)
    sinp = sinp.clamp(min=-1.0, max=1.0)
    pitch = sinp.asin()

    siny_cosp = 2.0 * (w * z + x * y)
    cosy_cosp = 1.0 - 2.0 * (yy + z * z)
    yaw = siny_cosp.atan2(cosy_cosp)

    return roll, pitch, yaw


def quaternion_from_euler(
    roll: Tensor, pitch: Tensor, yaw: Tensor
) -> tuple[Tensor, Tensor, Tensor, Tensor]:
    """
    Convert Euler angles to quaternion coefficients.

    Euler angles are assumed to be in radians in XYZ convention.

    Parameters
    ----------
    roll: Tensor
        The roll euler angle.
    pitch: Tensor
        The pitch euler angle.
    yaw: Tensor
        The yaw euler angle.

    Returns
    -------
    tuple[Tensor, Tensor, Tensor, Tensor]
        A tuple with quaternion coefficients in order of `wxyz`.

    Examples
    --------
    >>> roll, pitch, yaw = torch.tensor(0.0), torch.tensor(0.0), torch.tensor(0.0)
    >>> w, x, y, z = quaternion_from_euler(roll, pitch, yaw)
    >>> print(w, x, y, z)
    tensor(1.) tensor(0.) tensor(0.) tensor(0.)
    """
    roll_half = roll * 0.5
    pitch_half = pitch * 0.5
    yaw_half = yaw * 0.5

    cy = yaw_half.cos()
    sy = yaw_half.sin()
    cp = pitch_half.cos()
    sp = pitch_half.sin()
    cr = roll_half.cos()
    sr = roll_half.sin()

    qw = cy * cp * cr + sy * sp * sr
    qx = cy * cp * sr - sy * sp * cr
    qy = sy * cp * sr + cy * sp * cr
    qz = sy * cp * cr - cy * sp * sr

    return qw, qx, qy, qz


def normalize_pixel_coordinates(
    pixel_coordinates: Tensor, height: int, width: int, eps: float = 1e-8
) -> Tensor:
    """
    Normalize pixel coordinates between -1 and 1.

    Normalized, -1 if on extreme left, 1 if on extreme right (x = w-1).

    Parameters
    ----------
    pixel_coordinates: Tensor
        The grid with pixel coordinates. Size can be :math:`(*, 2)`.
    width: int
        The maximum width in the x-axis.
    height: int
        The maximum height in the y-axis.
    eps: float
        Safe division by zero.

    Returns
    -------
    Tensor[*, 2]
        Normalized pixel coordinates with shape :math:`(*, 2)`.

    Examples
    --------
    >>> pixel_coordinates = torch.tensor([[0.0, 0.0]])
    >>> normalized_coordinates = normalize_pixel_coordinates(pixel_coordinates, 100, 50)
    >>> print(normalized_coordinates)
    tensor([[-1., -1.]])
    """
    dtype = pixel_coordinates.dtype
    with pixel_coordinates.device:
        hw: Tensor = torch.stack(
            [
                torch.tensor(width, dtype=dtype),
                torch.tensor(height, dtype=dtype),
            ]
        )
        factor: Tensor = torch.tensor(2.0, dtype=dtype) / (hw - 1).clamp(eps)

    return factor * pixel_coordinates - 1


def denormalize_pixel_coordinates(
    pixel_coordinates: Tensor, height: int, width: int, eps: float = 1e-8
) -> Tensor:
    """
    Denormalize pixel coordinates.

    The input is assumed to be -1 if on extreme left, 1 if on extreme right (x = w-1).

    Parameters
    ----------
    pixel_coordinates: Tensor
        The normalized grid coordinates. Size can be :math:`(*, 2)`.
    width: int
        The maximum width in the x-axis.
    height: int
        The maximum height in the y-axis.
    eps: float
        Safe division by zero.

    Returns
    -------
    Tensor[*, 2]
        The denormalized pixel coordinates.

    Examples
    --------
    >>> coords = torch.tensor([[-1.0, -1.0]])
    >>> denormalized_coords = denormalize_pixel_coordinates(coords, 100, 50)
    >>> print(denormalized_coords)
    tensor([[0., 0.]])
    """
    dtype = pixel_coordinates.dtype
    with pixel_coordinates.device:
        hw: Tensor = torch.stack(
            [torch.tensor(width, dtype=dtype), torch.tensor(height, dtype=dtype)]
        )

    factor: Tensor = torch.tensor(2.0) / (hw - 1).clamp(eps)

    return 1.0 / factor * (pixel_coordinates + 1)


def normalize_pixel_coordinates3d(
    pixel_coordinates: Tensor, depth: int, height: int, width: int, eps: float = 1e-8
) -> Tensor:
    """
    Normalize pixel coordinates between -1 and 1.

    Normalized, -1 if on extreme left, 1 if on extreme right (x = w-1).

    Parameters
    ----------
    pixel_coordinates: Tensor
        The grid with pixel coordinates. Size can be :math:`(*, 3)`.
    depth: int
        The maximum depth in the z-axis.
    height: int
        The maximum height in the y-axis.
    width: int
        The maximum width in the x-axis.
    eps: float
        Safe division by zero.

    Returns
    -------
    Tensor[*, 3]
        The normalized pixel coordinates.

    Examples
    --------
    >>> pixel_coordinates = torch.tensor([[0.0, 0.0, 0.0]])
    >>> normalized_coordinates = normalize_pixel_coordinates3d(
    ...     pixel_coordinates, 10, 100, 50
    ... )
    >>> print(normalized_coordinates)
    tensor([[-1., -1., -1.]])
    """
    if pixel_coordinates.shape[-1] != 3:
        msg = f"Input pixel_coordinates must be of shape (*, 3). Got {pixel_coordinates.shape}"
        raise ValueError(msg)
    # compute normalization factor
    dhw: Tensor = (
        torch.stack([torch.tensor(depth), torch.tensor(width), torch.tensor(height)])
        .to(pixel_coordinates.device)
        .to(pixel_coordinates.dtype)
    )

    factor: Tensor = 2.0 / (dhw - 1).clamp(eps)

    return factor * pixel_coordinates - 1


def denormalize_pixel_coordinates3d(
    pixel_coordinates: Tensor, depth: int, height: int, width: int, eps: float = 1e-8
) -> Tensor:
    """
    Denormalize pixel coordinates.

    The input is assumed to be -1 if on extreme left, 1 if on extreme right (x = w-1).

    Parameters
    ----------
    pixel_coordinates: Tensor
        The normalized grid coordinates. Size can be :math:`(*, 3)`.
    depth: int
        The maximum depth in the z-axis.
    height: int
        The maximum height in the y-axis.
    width: int
        The maximum width in the x-axis.
    eps: float
        Safe division by zero.

    Returns
    -------
    Tensor[*, 3]
        The denormalized pixel coordinates.

    Examples
    --------
    >>> coords = torch.tensor([[-1.0, -1.0, -1.0]])
    >>> denormalized_coords = denormalize_pixel_coordinates3d(coords, 10, 100, 50)
    >>> print(denormalized_coords)
    tensor([[0., 0., 0.]])
    """
    dtype = pixel_coordinates.dtype
    with pixel_coordinates.device:
        dhw = torch.stack(
            [
                torch.tensor(depth, dtype=dtype),
                torch.tensor(width, dtype=dtype),
                torch.tensor(height, dtype=dtype),
            ]
        )
    factor: Tensor = torch.tensor(2.0) / (dhw - 1).clamp(eps)

    return torch.tensor(1.0) / factor * (pixel_coordinates + 1)


def angle_to_rotation(angle: Tensor) -> Tensor:
    """
    Create a rotation matrix out of angles in degrees.

    Parameters
    ----------
    angle: Tensor
        Tensor of angles in degrees, any shape :math:`(*)`.

    Returns
    -------
    Tensor[*, 2, 2]
        Tensor of rotation matrices.

    Examples
    --------
    >>> input = torch.tensor([45.0])
    >>> rotation_matrix = angle_to_rotation(input)
    >>> print(rotation_matrix)
    tensor([[[ 0.7071,  0.7071],
             [-0.7071,  0.7071]]])
    """
    ang_rad = deg2rad(angle)
    cos_a: Tensor = torch.cos(ang_rad)
    sin_a: Tensor = torch.sin(ang_rad)
    return torch.stack([cos_a, sin_a, -sin_a, cos_a], dim=-1).view(*angle.shape, 2, 2)


def normalize_homography(
    dst_pix_trans_src_pix: Tensor,
    dsize_src: tuple[int, int],
    dsize_dst: tuple[int, int],
) -> Tensor:
    """
    Normalize a given homography in pixels to [-1, 1].

    Parameters
    ----------
    dst_pix_trans_src_pix: Tensor[*, 3, 3]
        Homography/ies from source to destination to be normalized.
    dsize_src: tuple[int, int]
        Size of the source image (height, width).
    dsize_dst: tuple[int, int]
        Size of the destination image (height, width).

    Returns
    -------
    Tensor[*, 3, 3]
        The normalized homography.

    Examples
    --------
    >>> dst_pix_trans_src_pix = torch.eye(3)
    >>> dsize_src = (100, 100)
    >>> dsize_dst = (200, 200)
    >>> normalized_homography = normalize_homography(
    ...     dst_pix_trans_src_pix, dsize_src, dsize_dst
    ... )
    >>> print(normalized_homography)
    tensor([[ 2.0000,  0.0000, -1.0000],
            [ 0.0000,  2.0000, -1.0000],
            [ 0.0000,  0.0000,  1.0000]])
    """
    if not isinstance(dst_pix_trans_src_pix, torch.Tensor):
        msg = f"Input type is not a Tensor. Got {type(dst_pix_trans_src_pix)}"
        raise TypeError(msg)

    if not (
        len(dst_pix_trans_src_pix.shape) == 3
        or dst_pix_trans_src_pix.shape[-2:] == (3, 3)
    ):
        msg = f"Input dst_pix_trans_src_pix must be a Bx3x3 tensor. Got {dst_pix_trans_src_pix.shape}"
        raise ValueError(msg)

    # source and destination sizes
    src_h, src_w = dsize_src
    dst_h, dst_w = dsize_dst

    # compute the transformation pixel/norm for src/dst
    src_norm_trans_src_pix: Tensor = normal_transform_pixel(src_h, src_w).to(
        dst_pix_trans_src_pix
    )

    src_pix_trans_src_norm = unsafe_inverse(src_norm_trans_src_pix)
    dst_norm_trans_dst_pix: Tensor = normal_transform_pixel(dst_h, dst_w).to(
        dst_pix_trans_src_pix
    )

    # compute chain transformations
    dst_norm_trans_src_norm: Tensor = dst_norm_trans_dst_pix @ (
        dst_pix_trans_src_pix @ src_pix_trans_src_norm
    )
    return dst_norm_trans_src_norm


def normal_transform_pixel(
    height: int,
    width: int,
    eps: float = 1e-14,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
) -> Tensor:
    """
    Compute the normalization matrix from image size in pixels to [-1, 1].

    Parameters
    ----------
    height: int
        Image height.
    width: int
        Image width.
    eps: float
        Epsilon to prevent divide-by-zero errors.
    device: torch.device, optional
        Device to use.
    dtype: torch.dtype, optional
        Data type to use.

    Returns
    -------
    Tensor[1, 3, 3]
        Normalized transform.

    Examples
    --------
    >>> height, width = 100, 100
    >>> transform = normal_transform_pixel(height, width)
    >>> print(transform)
    tensor([[[ 0.0200,  0.0000, -1.0000],
             [ 0.0000,  0.0200, -1.0000],
             [ 0.0000,  0.0000,  1.0000]]])
    """
    tr_mat = torch.tensor(
        [[1.0, 0.0, -1.0], [0.0, 1.0, -1.0], [0.0, 0.0, 1.0]],
        device=device,
        dtype=dtype,
    )  # 3x3

    # prevent divide by zero bugs
    width_denom: float = eps if width == 1 else width - 1.0
    height_denom: float = eps if height == 1 else height - 1.0

    tr_mat[0, 0] = tr_mat[0, 0] * 2.0 / width_denom
    tr_mat[1, 1] = tr_mat[1, 1] * 2.0 / height_denom

    return tr_mat.unsqueeze(0)  # 1x3x3


def normal_transform_pixel3d(
    depth: int,
    height: int,
    width: int,
    eps: float = 1e-14,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
) -> Tensor:
    """
    Compute the normalization matrix from image size in pixels to [-1, 1].

    Parameters
    ----------
    depth: int
        Image depth.
    height: int
        Image height.
    width: int
        Image width.
    eps: float
        Epsilon to prevent divide-by-zero errors.
    device: torch.device, optional
        Device to use.
    dtype: torch.dtype, optional
        Data type to use.

    Returns
    -------
    Tensor[1, 4, 4]
        Normalized transform.

    Examples
    --------
    >>> depth, height, width = 10, 100, 100
    >>> transform = normal_transform_pixel3d(depth, height, width)
    >>> print(transform)
    tensor([[[ 0.2000,  0.0000,  0.0000, -1.0000],
             [ 0.0000,  0.0200,  0.0000, -1.0000],
             [ 0.0000,  0.0000,  0.0200, -1.0000],
             [ 0.0000,  0.0000,  0.0000,  1.0000]]])
    """
    tr_mat = unsafe_inverse(
        torch.tensor(
            [
                [1.0, 0.0, 0.0, -1.0],
                [0.0, 1.0, 0.0, -1.0],
                [0.0, 0.0, 1.0, -1.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            device=device,
            dtype=dtype,
        )
    )  # 4x4

    # prevent divide by zero bugs
    width_denom: float = eps if width == 1 else width - 1.0
    height_denom: float = eps if height == 1 else height - 1.0
    depth_denom: float = eps if depth == 1 else depth - 1.0

    tr_mat[0, 0] = tr_mat[0, 0] * 2.0 / width_denom
    tr_mat[1, 1] = tr_mat[1, 1] * 2.0 / height_denom
    tr_mat[2, 2] = tr_mat[2, 2] * 2.0 / depth_denom

    return tr_mat.unsqueeze(0)  # 1x4x4


def denormalize_homography(
    dst_pix_trans_src_pix: Tensor,
    dsize_src: tuple[int, int],
    dsize_dst: tuple[int, int],
) -> Tensor:
    """
    De-normalize a given homography in pixels from [-1, 1] to actual height and width.

    Parameters
    ----------
    dst_pix_trans_src_pix: Tensor[*, 3, 3]
        Homography/ies from source to destination to be denormalized.
    dsize_src: tuple[int, int]
        Size of the source image (height, width).
    dsize_dst: tuple[int, int]
        Size of the destination image (height, width).

    Returns
    -------
    Tensor[*, 3, 3]
        The denormalized homography.

    Examples
    --------
    >>> dst_pix_trans_src_pix = torch.eye(3)
    >>> dsize_src = (100, 100)
    >>> dsize_dst = (200, 200)
    >>> denormalized_homography = denormalize_homography(
    ...     dst_pix_trans_src_pix, dsize_src, dsize_dst
    ... )
    >>> print(denormalized_homography)
    tensor([[ 0.5000,  0.0000,  0.5000],
            [ 0.0000,  0.5000,  0.5000],
            [ 0.0000,  0.0000,  1.0000]])
    """
    if not isinstance(dst_pix_trans_src_pix, torch.Tensor):
        msg = f"Input type is not a Tensor. Got {type(dst_pix_trans_src_pix)}"
        raise TypeError(msg)

    if not (
        len(dst_pix_trans_src_pix.shape) == 3
        or dst_pix_trans_src_pix.shape[-2:] == (3, 3)
    ):
        msg = f"Input dst_pix_trans_src_pix must be a Bx3x3 tensor. Got {dst_pix_trans_src_pix.shape}"
        raise ValueError(msg)

    # source and destination sizes
    src_h, src_w = dsize_src
    dst_h, dst_w = dsize_dst

    # compute the transformation pixel/norm for src/dst
    src_norm_trans_src_pix: Tensor = normal_transform_pixel(src_h, src_w).to(
        dst_pix_trans_src_pix
    )

    dst_norm_trans_dst_pix: Tensor = normal_transform_pixel(dst_h, dst_w).to(
        dst_pix_trans_src_pix
    )
    dst_denorm_trans_dst_pix = unsafe_inverse(dst_norm_trans_dst_pix)
    # compute chain transformations
    dst_norm_trans_src_norm: Tensor = dst_denorm_trans_dst_pix @ (
        dst_pix_trans_src_pix @ src_norm_trans_src_pix
    )
    return dst_norm_trans_src_norm


def normalize_homography3d(
    dst_pix_trans_src_pix: Tensor,
    dsize_src: tuple[int, int, int],
    dsize_dst: tuple[int, int, int],
) -> Tensor:
    """
    Normalize a given homography in pixels to [-1, 1].

    Parameters
    ----------
    dst_pix_trans_src_pix: Tensor[*, 4, 4]
        Homography/ies from source to destination to be normalized.
    dsize_src: tuple[int, int, int]
        Size of the source image (depth, height, width).
    dsize_dst: tuple[int, int, int]
        Size of the destination image (depth, height, width).

    Returns
    -------
    Tensor[*, 4, 4]
        The normalized homography.

    Examples
    --------
    >>> dst_pix_trans_src_pix = torch.eye(4)
    >>> dsize_src = (10, 100, 100)
    >>> dsize_dst = (20, 200, 200)
    >>> normalized_homography = normalize_homography3d(
    ...     dst_pix_trans_src_pix, dsize_src, dsize_dst
    ... )
    >>> print(normalized_homography)
    tensor([[ 2.0000,  0.0000,  0.0000, -1.0000],
            [ 0.0000,  2.0000,  0.0000, -1.0000],
            [ 0.0000,  0.0000,  2.0000, -1.0000],
            [ 0.0000,  0.0000,  0.0000,  1.0000]])
    """
    if not isinstance(dst_pix_trans_src_pix, torch.Tensor):
        msg = f"Input type is not a Tensor. Got {type(dst_pix_trans_src_pix)}"
        raise TypeError(msg)

    if not (
        len(dst_pix_trans_src_pix.shape) == 3
        or dst_pix_trans_src_pix.shape[-2:] == (4, 4)
    ):
        msg = f"Input dst_pix_trans_src_pix must be a Bx3x3 tensor. Got {dst_pix_trans_src_pix.shape}"
        raise ValueError(msg)

    # source and destination sizes
    src_d, src_h, src_w = dsize_src
    dst_d, dst_h, dst_w = dsize_dst
    # compute the transformation pixel/norm for src/dst
    src_norm_trans_src_pix: Tensor = normal_transform_pixel3d(src_d, src_h, src_w).to(
        dst_pix_trans_src_pix
    )

    src_pix_trans_src_norm = unsafe_inverse(src_norm_trans_src_pix)
    dst_norm_trans_dst_pix: Tensor = normal_transform_pixel3d(dst_d, dst_h, dst_w).to(
        dst_pix_trans_src_pix
    )
    # compute chain transformations
    dst_norm_trans_src_norm: Tensor = dst_norm_trans_dst_pix @ (
        dst_pix_trans_src_pix @ src_pix_trans_src_norm
    )
    return dst_norm_trans_src_norm


def normalize_points_with_intrinsics(point_2d: Tensor, camera_matrix: Tensor) -> Tensor:
    """
    Normalizes points with intrinsics. Useful for conversion of keypoints to be used with essential matrix.

    Parameters
    ----------
    point_2d: Tensor[*, 2]
        Points in the image pixel coordinates. The shape of the tensor can be :math:`(*, 2)`.
    camera_matrix: Tensor[*, 3, 3] or Tensor[*, 4, 4]
        Tensor containing the intrinsics camera matrix.

    Returns
    -------
    Tensor[*, 2]
        Tensor of (u, v) cam coordinates.

    Examples
    --------
    >>> point_2d = torch.tensor([[100.0, 200.0]])
    >>> camera_matrix = torch.tensor(
    ...     [[1000.0, 0.0, 500.0], [0.0, 1000.0, 500.0], [0.0, 0.0, 1.0]]
    ... )
    >>> normalized_points = normalize_points_with_intrinsics(point_2d, camera_matrix)
    >>> print(normalized_points)
    tensor([[-0.4000, -0.3000]])
    """
    # projection eq. K_inv * [u v 1]'
    # x = (u - cx) * Z / fx
    # y = (v - cy) * Z / fy

    # unpack coordinates
    u_coord: Tensor = point_2d[..., 0]
    v_coord: Tensor = point_2d[..., 1]

    # unpack intrinsics
    fx: Tensor = camera_matrix[..., 0, 0]
    fy: Tensor = camera_matrix[..., 1, 1]
    cx: Tensor = camera_matrix[..., 0, 2]
    cy: Tensor = camera_matrix[..., 1, 2]

    # projective
    x_coord: Tensor = (u_coord - cx) / fx
    y_coord: Tensor = (v_coord - cy) / fy

    xy: Tensor = torch.stack([x_coord, y_coord], dim=-1)
    return xy


def denormalize_points_with_intrinsics(
    point_2d_norm: Tensor, camera_matrix: Tensor
) -> Tensor:
    """
    Normalizes points with intrinsics.
    Useful for conversion of keypoints to be used with essential matrix.

    Parameters
    ----------
    point_2d_norm: Tensor[*, 2]
        Tensor containing the 2d points in the image pixel coordinates. The shape of the tensor can be :math:`(*, 2)`.
    camera_matrix: Tensor[*, 3, 3] or Tensor[*, 4, 4]
        Tensor containing the intrinsics camera matrix.

    Returns
    -------
    Tensor[*, 2]
        Tensor of (u, v) cam coordinates.

    Examples
    --------
    >>> point_2d_norm = torch.tensor([[-0.4000, -0.3000]])
    >>> camera_matrix = torch.tensor(
    ...     [[1000.0, 0.0, 500.0], [0.0, 1000.0, 500.0], [0.0, 0.0, 1.0]]
    ... )
    >>> denormalized_points = denormalize_points_with_intrinsics(
    ...     point_2d_norm, camera_matrix
    ... )
    >>> print(denormalized_points)
    tensor([[100., 200.]])
    """
    # projection eq. [u, v, w]' = K * [x y z 1]'
    # u = fx * X + cx
    # v = fy * Y + cy

    # unpack coordinates
    x_coord: Tensor = point_2d_norm[..., 0]
    y_coord: Tensor = point_2d_norm[..., 1]

    # unpack intrinsics
    fx: Tensor = camera_matrix[..., 0, 0]
    fy: Tensor = camera_matrix[..., 1, 1]
    cx: Tensor = camera_matrix[..., 0, 2]
    cy: Tensor = camera_matrix[..., 1, 2]

    # apply intrinsics ans return
    u_coord: Tensor = x_coord * fx + cx
    v_coord: Tensor = y_coord * fy + cy

    return torch.stack([u_coord, v_coord], dim=-1)


def motion_to_extrinsics(R: Tensor, t: Tensor) -> Tensor:
    """
    Convert rotation matrix and translation vector to extrinsics matrix.

    Parameters
    ----------
    R: Tensor[*, 3, 3]
        Rotation matrix.
    t: Tensor[*, 3, 1]
        Translation vector.

    Returns
    -------
    Tensor[*, 4, 4]
        Extrinsic matrix.

    Examples
    --------
    >>> R = torch.eye(3)
    >>> t = torch.tensor([[0.0], [0.0], [0.0]])
    >>> extrinsics = motion_to_extrinsics(R, t)
    >>> print(extrinsics)
    tensor([[1., 0., 0., 0.],
            [0., 1., 0., 0.],
            [0., 0., 1., 0.],
            [0., 0., 0., 1.]])
    """
    Rt = torch.cat([R, t], dim=2)
    return convert_affinematrix_to_homography(Rt)


def extrinsics_to_motion(extrinsics: Tensor) -> tuple[Tensor, Tensor]:
    """
    Extracts rotation matrix and translation vector from extrinsics matrix.

    Parameters
    ----------
    extrinsics: Tensor[*, 4, 4]
        Pose matrix.

    Returns
    -------
    tuple[Tensor, Tensor]
        - Rotation matrix, :math:`(*, 3, 3)`.
        - Translation matrix :math:`(*, 3, 1)`.

    Examples
    --------
    >>> extrinsics = torch.eye(4)
    >>> R, t = extrinsics_to_motion(extrinsics)
    >>> print(R)
    tensor([[1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]])
    >>> print(t)
    tensor([[0.],
            [0.],
            [0.]])
    """
    return extrinsics[..., :3, :3], extrinsics[..., :3, 3:]


def reverse_motion(R: Tensor, t: Tensor) -> tuple[Tensor, Tensor]:
    """
    Reverses a camera-to-world transformation to world-to-camera transformation, or
    vice versa.

    Parameters
    ----------
    R: Tensor[*, 3, 3]
        Rotation matrix.
    t: Tensor[*, 3, 1]
        Translation matrix.

    Returns
    -------
    tuple[Tensor, Tensor]
        - Rotation matrix, :math:`(*, 3, 3)`.
        - Translation matrix :math:`(*, 3, 1)`.

    Examples
    --------
    >>> R = torch.eye(3)
    >>> t = torch.tensor([[0.0], [0.0], [0.0]])
    >>> R_inv, t_inv = reverse_motion(R, t)
    >>> print(R_inv)
    tensor([[1., 0., 0.],
            [0., 1., 0.],
            [0., 0., 1.]])
    >>> print(t_inv)
    tensor([[0.],
            [0.],
            [0.]])
    """
    R_inv = R.transpose(1, 2)
    new_t: Tensor = -R_inv @ t

    return (R_inv, new_t)


def reverse_extrinsics(extrinsics: Tensor) -> Tensor:
    """
    Reverses a camera-to-world transformation to world-to-camera transformation, or
    vice versa.

    Parameters
    ----------
    extrinsics: Tensor[*, 4, 4]
        Pose matrix.

    Returns
    -------
    Tensor[*, 4, 4]
        Pose matrix.

    Examples
    --------
    >>> extrinsics = torch.eye(4)
    >>> reversed_extrinsics = reverse_extrinsics(extrinsics)
    >>> print(reversed_extrinsics)
    tensor([[1., 0., 0., 0.],
            [0., 1., 0., 0.],
            [0., 0., 1., 0.],
            [0., 0., 0., 1.]])
    """
    return motion_to_extrinsics(*reverse_motion(*extrinsics_to_motion(extrinsics)))


def vector_to_skew_symmetric_matrix(vec: Tensor) -> Tensor:
    r"""
    Converts a vector to a skew symmetric matrix.

    A vector :math:`(v1, v2, v3)` has a corresponding skew-symmetric matrix, which is of the form:

    .. math::
        \begin{bmatrix} 0 & -v3 & v2 \\
        v3 & 0 & -v1 \\
        -v2 & v1 & 0\\end{bmatrix}

    Parameters
    ----------
    vec: Tensor[*, 3]
        Input vector.

    Returns
    -------
    Tensor[*, 3, 3]
        The skew-symmetric matrix.

    Examples
    --------
    >>> vec = torch.tensor([1.0, 2.0, 3.0])
    >>> skew_symmetric_matrix = vector_to_skew_symmetric_matrix(vec)
    >>> print(skew_symmetric_matrix)
    tensor([[ 0., -3.,  2.],
            [ 3.,  0., -1.],
            [-2.,  1.,  0.]])
    """
    v1, v2, v3 = vec[..., 0], vec[..., 1], vec[..., 2]
    zeros = torch.zeros_like(v1)
    return torch.stack(
        [
            torch.stack([zeros, -v3, v2], dim=-1),
            torch.stack([v3, zeros, -v1], dim=-1),
            torch.stack([-v2, v1, zeros], dim=-1),
        ],
        dim=-2,
    )


####################################################
# Various helper methods for geometric projections #
####################################################


def apply_points(transform: Tensor, points: Tensor) -> Tensor:
    """
    Apply a transformation on a set of points.

    Parameters
    ----------
    transform: Tensor[..., D+1, D+1]
        Transformation matrix.
    points: Tensor[..., N, D]
        Points to transform.

    Returns
    -------
    Tensor[*, N, D]
        Transformed points.

    Examples
    --------
    >>> transform = torch.eye(4)
    >>> points = torch.tensor([[1.0, 2.0, 3.0]])
    >>> transformed_points = apply_points(transform, points)
    >>> print(transformed_points)
    tensor([[1., 2., 3.]])
    """
    batch_size = points.shape[:-2]
    # Accept variable number of dimensions. This reshape operation combines
    # both unsqueezing (in case of no batch) and flattening (in case of multiple
    # leading dimensions).
    points = points.reshape(-1, points.shape[-2], points.shape[-1])
    transform = transform.reshape(-1, transform.shape[-2], transform.shape[-1])

    # Repeat the transform matrix (which may be unit) to match the number of points
    transform = torch.repeat_interleave(
        transform, repeats=points.shape[0] // transform.shape[0], dim=0
    )

    points_homo = euclidean_to_homogeneous_points(points)

    result_homo = torch.bmm(points_homo, transform.permute(0, 2, 1))
    result_homo = torch.squeeze(result_homo, dim=-1)
    result = homogeneous_to_euclidean_points(result_homo)
    if len(batch_size) > 0:
        return result.unflatten(0, batch_size)
    return result.squeeze(0)


# ------------------------------- #
# Spherical coordinate conversion #
# ------------------------------- #


def spherical_zbuffer_to_euclidean(spherical_tensor: Tensor) -> Tensor:
    """
    Convert spherical coordinates with z-buffer depth to Euclidean coordinates.

    Parameters
    ----------
    spherical_tensor: Tensor
        Spherical coordinates tensor.

    Returns
    -------
    Tensor
        Euclidean coordinates tensor.

    Examples
    --------
    >>> spherical_tensor = torch.tensor([[1.5708, 1.5708, 1.0]])
    >>> euclidean_tensor = spherical_zbuffer_to_euclidean(spherical_tensor)
    >>> print(euclidean_tensor)
    tensor([[1.0000, 0.0000, 1.0000]])
    """
    theta, phi, z = spherical_tensor.unbind(-1)
    x = z * torch.tan(theta)
    y = z / torch.tan(phi) / torch.cos(theta)
    return _stack_3_last(x, y, z)


def spherical_to_euclidean(spherical_tensor: Tensor) -> Tensor:
    """
    Convert spherical coordinates to Euclidean coordinates.

    Parameters
    ----------
    spherical_tensor: Tensor
        Spherical coordinates tensor.

    Returns
    -------
    Tensor
        Euclidean coordinates tensor.

    Examples
    --------
    >>> spherical_tensor = torch.tensor([[1.5708, 1.5708, 1.0]])
    >>> euclidean_tensor = spherical_to_euclidean(spherical_tensor)
    >>> print(euclidean_tensor)
    tensor([[1.0000, 0.0000, 0.0000]])
    """
    theta, phi, r = spherical_tensor.unbind(-1)
    x = r * torch.sin(phi) * torch.sin(theta)
    y = r * torch.cos(phi)
    z = r * torch.cos(theta) * torch.sin(phi)
    return _stack_3_last(x, y, z)


def euclidean_to_spherical(spherical_tensor: Tensor) -> Tensor:
    """
    Convert Euclidean coordinates to spherical coordinates.

    Parameters
    ----------
    spherical_tensor: Tensor
        Euclidean coordinates tensor.

    Returns
    -------
    Tensor
        Spherical coordinates tensor.

    Examples
    --------
    >>> euclidean_tensor = torch.tensor([[1.0, 0.0, 0.0]])
    >>> spherical_tensor = euclidean_to_spherical(euclidean_tensor)
    >>> print(spherical_tensor)
    tensor([[1.5708, 1.5708, 1.0000]])
    """
    x, y, z = spherical_tensor.unbind(-1)
    r = torch.sqrt(x**2 + y**2 + z**2)
    theta = torch.atan2(x / r, z / r)
    phi = torch.acos(y / r)

    return _stack_3_last(theta, phi, r)


def euclidean_to_spherical_zbuffer(euclidean_tensor: Tensor) -> Tensor:
    """
    Convert Euclidean coordinates to spherical coordinates with z-buffer depth.

    Parameters
    ----------
    euclidean_tensor: Tensor
        Euclidean coordinates tensor.

    Returns
    -------
    Tensor
        Spherical coordinates tensor.

    Examples
    --------
    >>> euclidean_tensor = torch.tensor([[1.0, 0.0, 1.0]])
    >>> spherical_tensor = euclidean_to_spherical_zbuffer(euclidean_tensor)
    >>> print(spherical_tensor)
    tensor([[0.0000, 1.5708, 1.0000]])
    """
    pitch = torch.asin(euclidean_tensor[..., 1])
    yaw = torch.atan2(euclidean_tensor[..., 0], euclidean_tensor[..., -1])
    z = euclidean_tensor[..., 2]
    return _stack_3_last(pitch, yaw, z)


# --------------------- #
# FOV-to-coordinate map #
# --------------------- #


def generate_fovmap(
    focal_length: Tensor, principal_point: Tensor, like: Size | Tensor
) -> Tensor:
    """Encode the camera intrinsic parameters as a 2D map.

    Uses (focal length and principle point) to generate a map with
    with channels for the distance to the principle point and the field of view.

    Parameters
    ----------
    focal_length: Tensor
        Focal length of the camera.
    principal_point: Tensor
        Principal point of the camera.
    like: Size or Tensor
        Size of the output map.

    Returns
    -------
    Tensor
        Field of view map.

    Examples
    --------
    >>> focal_length = torch.tensor([1000.0, 1000.0])
    >>> principal_point = torch.tensor([500.0, 500.0])
    >>> like = (100, 100)
    >>> fovmap = generate_fovmap(focal_length, principal_point, like)
    >>> print(fovmap.shape)
    torch.Size([100, 100, 4])
    """
    if isinstance(like, torch.Tensor):
        height, width = like.shape[-2:]
    else:
        height, width = like[-2:]

    batch_shape = focal_length.shape[:-1]
    assert principal_point.shape == focal_length.shape

    if len(batch_shape) == 0:
        principal_point = principal_point.unsqueeze(0)
        focal_length = focal_length.unsqueeze(0)
    elif len(batch_shape) > 1:
        focal_length = focal_length.flatten(0, -2)
        principal_point = principal_point.flatten(0, -2)

    device = focal_length.device
    dtype = focal_length.dtype if focal_length.is_floating_point() else torch.float32

    results = []
    for focal_length, principal_point in zip(
        focal_length, principal_point, strict=False
    ):
        fx, fy = focal_length.unbind(-1)
        u0, v0 = principal_point.unbind(-1)

        # Distance to principal point
        x_row = torch.arange(0, width, dtype=dtype, device=device)
        x_row_center_norm = (x_row - u0) / (width * 2)
        x_center = torch.tile(x_row_center_norm, (height, 1))  # [H, W]

        y_col = torch.arange(0, height, dtype=dtype, device=device)
        y_col_center_norm = (y_col - v0) / (height * 2)
        y_center = torch.tile(y_col_center_norm, (width, 1)).T

        # FoV
        fov_x = torch.arctan(x_center / (fx / width))
        fov_y = torch.arctan(y_center / (fy / height))

        res = torch.stack([x_center, y_center, fov_x, fov_y], dim=2)
        results.append(res)

    res = torch.stack(results)

    return res.reshape(*batch_shape, height, width, 4)


# --------------------------- #
# Generating coordinate grids #
# --------------------------- #


class GridMode(E.StrEnum):
    """Enumerates grid generation modes."""

    INDEX = E.auto()  # Return indices 0, ..., |G| - 1
    PIXEL_CENTER = E.auto()  # Return center pixels 0.5, ..., |G| - 0.5
    PIXEL_NOISE = E.auto()  # Return center pixels with additive uncertainty ~ 1


type GridType = GridMode | typing.Literal["index", "pixel_center", "pixel_noise"]


def generate_coord_grid_like(
    tensor: Tensor,
    mode: GridMode | str = "index",
    *,
    stride: int = 1,
) -> Tensor:
    r"""
    Generate pixel coordinates grid like the input tensor.

    Parameters
    ----------
    tensor:
        Input tensor.
    stride:
        The stride of the input tensor with respect to the canvas size.
    mode:
        Grid mode to use. By default, `GridMode.INDEX`.

    Returns
    -------
    Tensor[H, W, 2]
        Pixel coordinates grid.
    """
    H, W = tensor.shape[-2:]
    x_i = _arange_like(tensor, W * stride)
    y_i = _arange_like(tensor, H * stride)
    return _grid_from_indices(x_i, y_i, mode)


def generate_coord_grid_as(
    tensor: Tensor,
    canvas_size: Size | tuple[int, int],
    *,
    mode: GridMode | str = "index",
) -> Tensor:
    r"""
    Generate pixel coordinates grid with the same device and dtype as the given tensor.

    Parameters
    ----------
    tensor:
        Input tensor.
    canvas_size:
        Size of the canvas.
    mode:
        Grid mode to use. By default, `GridMode.INDEX`.

    Returns
    -------
    Tensor[H, W, 2]
        Pixel coordinates grid.
    """
    h, w = canvas_size
    x_idx = _arange_like(tensor, w)
    y_idx = _arange_like(tensor, h)
    return _grid_from_indices(x_idx, y_idx, mode)


def generate_coord_grid(
    canvas_size: torch.Size | tuple[int, int],
    mode: GridMode | str = "index",
    *,
    device: Device | None = None,
    dtype: DType | None = None,
) -> Tensor:
    r"""
    Generate pixel coordinates grid.

    Parameters
    ----------
    canvas_size:
        Size of the canvas.
    device:
        Device to use.
    dtype:
        Data type to use.
    mode:
        Grid mode to use. By default, `GridMode.INDEX`.

    Returns
    -------
    Tensor[H, W, 2]
        Pixel coordinates grid.
    """
    h, w = canvas_size
    if dtype is None:
        dtype = torch.float32
    x_idx = torch.arange(w, device=device, dtype=dtype)
    y_idx = torch.arange(h, device=device, dtype=dtype)
    return _grid_from_indices(x_idx, y_idx, mode)


def _grid_from_indices(
    x_i: Tensor,
    y_i: Tensor,
    mode: GridMode | str = "index",
) -> Tensor:
    # Generate a meshgrid of x and y coordinates
    # gx, gy = torch.broadcast_tensors(x.unsqueeze(0), y.unsqueeze(1))
    x_g, y_g = torch.meshgrid(x_i, y_i, indexing="xy")

    # Stack the x and y coordinates along the last dimension
    coords = _stack_2_last(x_g, y_g)

    # Apply the specified mode
    match mode:
        case GridMode.INDEX:
            pass
        case GridMode.PIXEL_CENTER:
            coords = coords + 0.5
        case GridMode.PIXEL_NOISE:
            coords += torch.rand_like(coords)
        case _:
            msg = f"Invalid {mode=}. Choose from {list(GridMode)}"
            raise ValueError(msg)
    return coords


# ------------------------------------ #
# Direction generation and conversions #
# ------------------------------------ #


def generate_directions_grid(
    transform: Tensor,
    canvas_size: tuple[int, int],
    *,
    dim: int | None = None,
    grid_mode: GridMode | str = GridMode.PIXEL_CENTER,
    normalize: bool = True,
    device: torch.device | None = None,
    dtype: torch.dtype | None = None,
) -> Tensor:
    """
    Generate directions for each pixel in the canvas.

    Parameters
    ----------
    transform: Tensor
        Transformation matrix.
    canvas_size: tuple[int, int]
        Size of the canvas.
    dim : int, optional
        Dimension where the grid will be created.
    grid_mode: GridMode
        Grid generator mode. By default pixel center.
    normalize: bool
        Whether to normalize the directions.
    device: torch.device, optional
        The device to generate the grid on.
    dtype: torch.dtype, optional
        The datatype of the grid.

    Returns
    -------
    Tensor
        Directions for each pixel.

    Examples
    --------
    >>> transform = torch.eye(4)
    >>> canvas_size = (100, 100)
    >>> directions = generate_directions(transform, canvas_size)
    >>> print(directions.shape)
    torch.Size([10000, 3])
    """
    # Generate the coordinate grid (H, W, 2)
    if device is None and dtype is None:
        coords = generate_coord_grid_as(transform, canvas_size, mode=grid_mode)
    else:
        if device is None:
            device = transform.device
        if dtype is None:
            dtype = transform.dtype
        coords = generate_coord_grid(
            canvas_size,
            device=device,
            dtype=dtype,
            mode=grid_mode,
        )

    # Sanitize coordinates
    coords = coords.flatten(0, 1)  # (H*W, 2)
    coords = euclidean_to_homogeneous_points(coords)  # (H*W, 3)
    coords = coords.repeat(transform.shape[0], 1, 1)  # (B, H*W, 3)

    # Inverse projection of each point
    proj = unsafe_inverse(transform)  # (B, 4, 4)
    dirs = apply_points(proj, coords)  # (B, H*W, 3)

    # Normalize directions (optional)
    if normalize:
        dirs = nn.functional.normalize(dirs.mT, dim=-2).mT

    # Reshape back to an image at the specified dimension
    if dim == 1:
        dirs = dirs.permute(0, 2, 1)  # (B, 3, H*W)
        dirs = dirs.unflatten(2, canvas_size)  # (B, 3, H, W)
    elif dim == 2:
        dirs = dirs.unflatten(2, canvas_size)  # (B, 3, H, W)
    elif dim is not None:
        msg = f"Expected {dim=} to be -1 (last dim), 1 (first dim) or None."
        raise ValueError(msg)

    return dirs


def directions_to_angles(dirs: Tensor, dim: int = -1) -> Tensor:
    """
    Convert directions to angles.

    Parameters
    ----------
    dirs: Tensor
        Directions tensor.

    Returns
    -------
    Tensor
        Angles tensor.

    Examples
    --------
    >>> dirs = torch.tensor([[1.0, 0.0, 0.0]])
    >>> angles = directions_to_angles(dirs)
    >>> print(angles)
    tensor([[1.5708, 1.5708]])
    """
    theta = torch.atan2(dirs[..., 0], dirs[..., -1])
    phi = torch.acos(dirs[..., 1])
    return _stack_2_last(theta, phi)


# ------------------- #
# Decomposition utils #
# ------------------- #


def _arange_like(tensor: Tensor, n: int) -> Tensor:
    return torch.arange(n, dtype=tensor.dtype, device=tensor.device)


torch.fx.wrap("_arange_like")  # type: ignore


def _stack_2_last(t0: Tensor, t1: Tensor) -> Tensor:
    """Stack 2 tensor in the last dimension."""
    return torch.stack((t0, t1), dim=-1)


def _stack_3_last(t0: Tensor, t1: Tensor, t2: Tensor) -> Tensor:
    """Stack 3 tensors in the last dimension."""
    return torch.stack((t0, t1, t2), dim=-1)


torch.fx.wrap("_stack_2_last")  # type: ignore
torch.fx.wrap("_stack_3_last")  # type: ignore



r"""Utilities for working with (boudning) boxes."""

from __future__ import annotations

import enum as E
import typing as T

import torch
from torchvision.tv_tensors import BoundingBoxFormat as _TVBBoxFormat


class BBoxFormat(E.StrEnum):
    """Coordinate format of a bounding box."""

    LTRB = E.auto()  # (left        , top       , right         , bottom)
    LTWH = E.auto()  # (left        , top       , width         , height)
    CCWH = E.auto()  # (center x    , center y  , width         , height)
    CCAH = E.auto()  # (center x    , center y  , aspect ratio  , height)


def translate_bbox_format_to_torchvision(src: BBoxFormat) -> _TVBBoxFormat:
    """
    Convert a :class:`BBoxFormat` to a :class:`torchvision.tv_tensors.BoundingBoxFormat`.

    Parameters
    ----------
    src : BBoxFormat
        The format to convert.

    Returns
    -------
    torchvision.tv_tensors.BoundingBoxFormat
        The converted format.

    Notes
    -----
    This method was added for completeness after torchvision introduced their own
    bounding box format. Users are encoraged to stick with the `BBoxFormat` wherever
    possible to avoid unnecessary conversions.
    """
    match src:
        case BBoxFormat.LTRB:
            dst = _TVBBoxFormat.XYXY
        case BBoxFormat.LTWH:
            dst = _TVBBoxFormat.XYWH
        case BBoxFormat.CCWH:
            dst = _TVBBoxFormat.CXCYWH
        case _:
            msg = f"Format {src!r} cannot be converted to a torchvision format"
            raise ValueError(msg)
    return dst


def translate_bbox_format_from_torchvision(
    src: _TVBBoxFormat,
) -> BBoxFormat:
    """
    Convert a :class:`torchvision.tv_tensors.BoundingBoxFormat` to a :class:`BBoxFormat`.

    Parameters
    ----------
    src : torchvision.tv_tensors.BoundingBoxFormat
        The format to convert.

    Returns
    -------
    BBoxFormat
        The converted format.

    Notes
    -----
    This method was added for completeness after torchvision introduced their own
    bounding box format. Users are encoraged to stick with the `BBoxFormat` wherever
    possible to avoid unnecessary conversions.
    """
    match src:
        case _TVBBoxFormat.XYXY:
            dst = BBoxFormat.LTRB
        case _TVBBoxFormat.XYWH:
            dst = BBoxFormat.LTWH
        case _TVBBoxFormat.CXCYWH:
            dst = BBoxFormat.CCWH
        case _:
            msg = f"Format {src!r} cannot be converted to a BBoxFormat"
            raise ValueError(msg)
    return dst


def convert_boxes(
    boxes: torch.Tensor,
    src: BBoxFormat | _TVBBoxFormat,
    dst: BBoxFormat | _TVBBoxFormat,
) -> torch.Tensor:
    """
    Convert a set of bounding boxes from one format to another.

    Parameters
    ----------
    boxes : torch.Tensor
        The boxes to convert.
    src : BBoxFormat
        The source (current) format of :attr:`boxes`.
    dst : BBoxFormat
        The destination (target) format of the converted boxes.

    Returns
    -------
    torch.Tensor
        The converted boxes.
    """
    if isinstance(src, _TVBBoxFormat):
        src = translate_bbox_format_from_torchvision(src)
    if isinstance(dst, _TVBBoxFormat):
        dst = translate_bbox_format_from_torchvision(dst)

    if src == dst:
        return boxes

    # Convert to ltrb
    match src:
        case BBoxFormat.LTWH:
            ltrb = ltwh_to_ltrb(boxes)
        case BBoxFormat.CCWH:
            ltrb = ccwh_to_ltrb(boxes)
        case BBoxFormat.CCAH:
            ltrb = ccah_to_ltrb(boxes)
        case _:
            msg = f"Unsupported conversion from {src!r}"
            raise NotImplementedError(msg)
    match dst:
        case BBoxFormat.LTRB:
            boxes = ltrb
        case BBoxFormat.LTWH:
            boxes = ltrb_to_ltwh(ltrb)
        case BBoxFormat.CCWH:
            boxes = ltrb_to_ccwh(ltrb)
        case BBoxFormat.CCAH:
            boxes = ltrb_to_ccah(ltrb)
        case _:
            msg = f"Unsupported conversion to {dst!r}"
            raise NotImplementedError(msg)

    return boxes


def boxes_to_areas(ltrb: torch.Tensor) -> torch.Tensor:
    """Compute the areas of a set of boxes."""
    if ltrb.numel() == 0:
        return torch.zeros(ltrb.shape[:-1], device=ltrb.device, dtype=ltrb.dtype)

    w = ltrb[..., 2] - ltrb[..., 0]
    h = ltrb[..., 3] - ltrb[..., 1]

    return w * h


def scale_boxes(
    ltrb: torch.Tensor,
    scale: torch.Tensor | torch.Size | float | T.Iterable[int | float],
) -> torch.Tensor:
    """
    Scale a set of boxes by a fixed ratio.

    Parameters
    ----------
    ltrb : torch.Tensor
        The boxes to scale, in [x1, y1, x2, y2] format.
    scale : Union[torch.Tensor, torch.Size, float, Sequence[float]]
        The scale factor(s) to apply to the boxes, when a single value is provided, it is
        used for both dimensions.

    Returns
    -------
    torch.Tensor
        The scaled boxes.
    """
    if ltrb.numel() == 0:
        return ltrb

    if isinstance(scale, (torch.Tensor)):
        sw, sh = map(float, scale.tolist())
    elif not isinstance(scale, T.Iterable):
        sw = sh = float(scale)
    else:
        sw, sh = map(float, scale)

    return torch.stack(
        [
            ltrb[..., 0] * sw,
            ltrb[..., 1] * sh,
            ltrb[..., 2] * sw,
            ltrb[..., 3] * sh,
        ],
        dim=-1,
    ).to(ltrb.dtype)


def reposition_boxes_in_crop(
    boxes_ltrb: torch.Tensor, crop_ltrb: torch.Tensor, clamp: bool = False
) -> torch.Tensor:
    """
    Reposition a set of boxes that have been defined relative to some image size to be
    relative to a crop of that image.

    Parameters
    ----------
    boxes_ltrb : torch.Tensor
        The boxes to reposition, in [x1, y1, x2, y2] format.
    crop_ltrb : torch.Tensor
        The crop to reposition the boxes into, in [x1, y1, x2, y2] format.
    clamp : bool
        Whether to clamp the boxes to the crop boundaries.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if boxes_ltrb.numel() == 0:
        return boxes_ltrb

    boxes_x1 = boxes_ltrb[..., 0]
    boxes_y1 = boxes_ltrb[..., 1]
    boxes_x2 = boxes_ltrb[..., 2]
    boxes_y2 = boxes_ltrb[..., 3]

    crop_x1 = crop_ltrb[..., 0]
    crop_y1 = crop_ltrb[..., 1]
    crop_x2 = crop_ltrb[..., 2]
    crop_y2 = crop_ltrb[..., 3]

    if clamp:
        boxes_x1 = torch.clamp(boxes_x1, min=crop_x1)
        boxes_y1 = torch.clamp(boxes_y1, min=crop_y1)
        boxes_x2 = torch.clamp(boxes_x2, max=crop_x2)
        boxes_y2 = torch.clamp(boxes_y2, max=crop_y2)

    boxes_x1 -= crop_x1
    boxes_y1 -= crop_y1
    boxes_x2 -= crop_x1
    boxes_y2 -= crop_y1

    return torch.stack([boxes_x1, boxes_y1, boxes_x2, boxes_y2], dim=-1)


def expand_boxes_to_squares(ltrb: torch.Tensor) -> torch.Tensor:
    """Pad a set of boxes such that they are square."""
    if ltrb.numel() == 0:
        return ltrb

    w = ltrb[..., 2] - ltrb[..., 0]
    h = ltrb[..., 3] - ltrb[..., 1]

    max_size = torch.max(w, h)
    half_diff = (max_size - torch.stack([w, h], dim=-1)) / 2

    return torch.stack(
        [
            ltrb[..., 0] - half_diff[..., 0],
            ltrb[..., 1] - half_diff[..., 1],
            ltrb[..., 2] + half_diff[..., 0],
            ltrb[..., 3] + half_diff[..., 1],
        ],
        dim=-1,
    )


def expand_boxes_by_pixels(ltrb: torch.Tensor, px: int) -> torch.Tensor:
    """
    Expand a set of boxes by a fixed number of pixels.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ltrb.numel() == 0:
        return ltrb

    return torch.stack(
        [
            ltrb[..., 0] - px,
            ltrb[..., 1] - px,
            ltrb[..., 2] + px,
            ltrb[..., 3] + px,
        ],
        dim=-1,
    )


def expand_boxes_by_ratio(ltrb: torch.Tensor, ratio: float) -> torch.Tensor:
    """
    Expand a set of boxes by a fixed ratio.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ltrb.numel() == 0:
        return ltrb

    w = ltrb[..., 2] - ltrb[..., 0]
    h = ltrb[..., 3] - ltrb[..., 1]

    half_diff_w = (w * ratio) / 2
    half_diff_h = (h * ratio) / 2

    return torch.stack(
        [
            ltrb[..., 0] - half_diff_w,
            ltrb[..., 1] - half_diff_h,
            ltrb[..., 2] + half_diff_w,
            ltrb[..., 3] + half_diff_h,
        ],
        dim=-1,
    )


def ltrb_to_ltwh(ltrb: torch.Tensor) -> torch.Tensor:
    """
    Convert boxes from [x1, y1, x2, y2] to [x, y, w, h] format.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ltrb.numel() == 0:
        return ltrb

    w = ltrb[..., 2] - ltrb[..., 0]
    h = ltrb[..., 3] - ltrb[..., 1]
    x = ltrb[..., 0]
    y = ltrb[..., 1]

    return torch.stack([x, y, w, h], dim=-1)


def ltwh_to_ltrb(ltwh: torch.Tensor) -> torch.Tensor:
    """
    Convert boxes from [x, y, w, h] to [x1, y1, x2, y2] format.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ltwh.numel() == 0:
        return ltwh

    x1 = ltwh[..., 0]
    y1 = ltwh[..., 1]
    x2 = x1 + ltwh[..., 2]
    y2 = y1 + ltwh[..., 3]

    return torch.stack([x1, y1, x2, y2], dim=-1)


def ltrb_to_ccwh(ltrb: torch.Tensor) -> torch.Tensor:
    """
    Convert boxes from [x1, y1, x2, y2] to [cx, cy, w, h] format.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ltrb.numel() == 0:
        return ltrb

    cx = (ltrb[..., 0] + ltrb[..., 2]) / 2
    cy = (ltrb[..., 1] + ltrb[..., 3]) / 2
    w = ltrb[..., 2] - ltrb[..., 0]
    h = ltrb[..., 3] - ltrb[..., 1]

    return torch.stack([cx, cy, w, h], dim=-1)


def ccwh_to_ltrb(ccwh: torch.Tensor) -> torch.Tensor:
    """
    Convert boxes from [cx, cy, w, h] to [x1, y1, x2, y2] format.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ccwh.numel() == 0:
        return ccwh

    x1 = ccwh[..., 0] - ccwh[..., 2] / 2
    y1 = ccwh[..., 1] - ccwh[..., 3] / 2
    x2 = ccwh[..., 0] + ccwh[..., 2] / 2
    y2 = ccwh[..., 1] + ccwh[..., 3] / 2

    return torch.stack([x1, y1, x2, y2], dim=-1)


def ltrb_to_ccah(ltrb: torch.Tensor) -> torch.Tensor:
    """
    Convert boxes from [x1, y1, x2, y2] to [cx, cy, ar, h] format.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ltrb.numel() == 0:
        return ltrb

    cx = (ltrb[..., 0] + ltrb[..., 2]) / 2
    cy = (ltrb[..., 1] + ltrb[..., 3]) / 2
    ar = (ltrb[..., 2] - ltrb[..., 0]) / (ltrb[..., 3] - ltrb[..., 1])
    h = ltrb[..., 3] - ltrb[..., 1]

    return torch.stack([cx, cy, ar, h], dim=-1)


def ccah_to_ltrb(ccah: torch.Tensor) -> torch.Tensor:
    """
    Convert boxes from [cx, cy, ar, h] to [x1, y1, x2, y2] format.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ccah.numel() == 0:
        return ccah

    x1 = ccah[..., 0] - ccah[..., 2] / 2
    y1 = ccah[..., 1] - ccah[..., 3] / 2
    x2 = ccah[..., 0] + ccah[..., 2] / 2
    y2 = ccah[..., 1] + ccah[..., 3] / 2

    return torch.stack([x1, y1, x2, y2], dim=-1)


def ccwh_to_ccah(ccwh: torch.Tensor) -> torch.Tensor:
    """
    Convert boxes from [cx, cy, w, h] to [cx, cy, ar, h] format.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ccwh.numel() == 0:
        return ccwh

    ar = ccwh[..., 2] / ccwh[..., 3]

    return torch.stack([ccwh[..., 0], ccwh[..., 1], ar, ccwh[..., 3]], dim=-1)


def ccah_to_ccwh(ccah: torch.Tensor) -> torch.Tensor:
    """
    Convert boxes from [cx, cy, ar, h] to [cx, cy, w, h] format.

    Notes
    -----
    This function was generated using GitHub Copilot.
    """
    if ccah.numel() == 0:
        return ccah

    w = ccah[..., 3] * ccah[..., 2]

    return torch.stack([ccah[..., 0], ccah[..., 1], w, ccah[..., 3]], dim=-1)



r"""
Checks for various conditions at runtime. Ignored at compile/fx/export.
The API is intended to be similar to those used in C++ extensions.
"""

import typing as T
from types import EllipsisType

import torch

from unipercept.types import Tensor


def assert_shape(  # noqa: C901
    x: Tensor, shape: T.Iterable[str | int | EllipsisType], *, raises: bool = True
) -> bool:
    r"""
    Check whether a tensor has a specified shape.

    The shape can be specified with a implicit or explicit list of strings.
    The guard also check whether the variable is a type `Tensor`.

    Parameters
    ----------
    x: Tensor
        The tensor to check.
    shape: Iterable[str | int | EllipsisType]
        The expected shape of the tensor. The ellipsis can be used to match any
        number of dimensions.
    raises: bool
        Whether to raise an exception if the shape does not match. By default, True.

    Raises
    ------
    TypeError: if the input tensor is not a tensor.

    Examples
    --------
    >>> x = torch.rand(2, 3, 8, 16)
    >>> assert_shape(x, ("B", "C", "H", "W"))
    True

    >>> x = torch.rand(2, 3, 8, 16)
    >>> assert_shape(x, ("B", "C", "H", "H"))
    False

    >>> x = torch.rand(2, 1, 1, 2, 1, 8, 16)
    >>> assert_shape(x, (..., "H", "W"))
    True

    >>> x = torch.rand(2, 3, 8, 16)
    >>> assert_shape(x, (2, 3, "H", "W"))
    True

    >>> x = torch.rand(2, 3, 8, 16)
    >>> assert_shape(x, (2, 3, "H", ...))
    True
    """
    if not assert_tensor(x, raises=raises):
        return False

    tgt = list(shape)
    src = list(x.shape)
    if tgt[0] == ...:
        src = x.shape[-len(tgt) + 1 :]
        tgt.pop(0)
    if tgt[-1] == ...:
        src = x.shape[: len(tgt) - 1]
        tgt.pop(-1)

    if any(dim is ... for dim in tgt):
        msg = f"Ellipsis must be at the beginning or end of the shape. Got {shape}"
        raise ValueError(msg)

    if len(src) != len(tgt):
        if raises:
            msg = f"{x} shape must be [{shape}]. Got {x.shape}"
            raise TypeError(msg)
        return False

    for i in range(len(src)):
        dim: str | int | EllipsisType = tgt[i]
        if isinstance(dim, str):
            continue
        if isinstance(dim, int):
            if src[i] != dim:
                if raises:
                    msg = f"{x} shape must be [{shape}]. Got {x.shape}"
                    raise AssertionError(msg)
                return False
        else:
            msg = (
                f"Invalid shape dimension. Got {dim=} for {shape=}. "
                "Checked dimensions must be a string or integer."
            )
            raise ValueError(msg)

    return True


def assert_tensor(x: object, *, raises: bool = True) -> T.TypeGuard[torch.Tensor]:
    r"""
    Check whether a variable is a tensor.

    Parameters
    ----------
    x: object
        The variable to check.
    raises: bool
        Whether to raise an exception if the variable is not a tensor. By default, True.

    Examples
    --------
    >>> x = torch.rand(2, 3, 8, 16)
    >>> assert_tensor(x)
    True

    >>> x = 5
    >>> assert_tensor(x)
    False
    """
    if not isinstance(x, torch.Tensor):
        if raises:
            msg = f"{x} must be a tensor. Got {type(x)}"
            raise AssertionError(msg)
        return False
    return True



"""Utility functions and classes."""

from unipercept.utils.python import defer_imports as _defer_imports

__all__ = [
    "box",
    "check",
    "descriptors",
    "flops",
    "formatter",
    "frozendict",
    "function",
    "image",
    "inspect",
    "latency",
    "mask",
    "missing",
    "pickle",
    "python",
    "signal",
    "string",
    "tensorclass",
    "timestamp",
    "ulid",
]

__getattr__, __dir__ = _defer_imports(__name__, __all__)



"""Various functional utilities."""

from __future__ import annotations

import collections.abc
import inspect
from collections.abc import Callable, Iterable, Sequence
from functools import partial, wraps
from itertools import repeat
from typing import (
    Any,
    Literal,
    ParamSpec,
    TypeVar,
    cast,
    overload,
)

import torch
from torch import Tensor
from torch._dynamo import allow_in_graph

__all__ = []

_Ps = ParamSpec("_Ps")

_R1 = TypeVar("_R1")
_R2 = TypeVar("_R2")
_R3 = TypeVar("_R3")
_R4 = TypeVar("_R4")
_R5 = TypeVar("_R5")
_R6 = TypeVar("_R6")
_R7 = TypeVar("_R7")
_R8 = TypeVar("_R8")


# Currently, the Python typing system doe not support casting each of the elements of a tuple to a different type,
# in our case a list, so we hardcode the first 9 cases.
@overload
def multi_apply[**Ps, R1, R2, R3, R4, R5, R6, R7, R8](
    func: Callable[_Ps, tuple[_R1, _R2, _R3, _R4, _R5, _R6, _R7, _R8]],
    *args: Iterable[Any],
    **kwargs: Any,
) -> tuple[
    list[_R1],
    list[_R2],
    list[_R3],
    list[_R4],
    list[_R5],
    list[_R6],
    list[_R7],
    list[_R8],
]: ...


@overload
def multi_apply[**Ps, R1, R2, R3, R4, R5, R6, R7](
    func: Callable[_Ps, tuple[_R1, _R2, _R3, _R4, _R5, _R6, _R7]],
    *args: Iterable[Any],
    **kwargs: Any,
) -> tuple[
    list[_R1], list[_R2], list[_R3], list[_R4], list[_R5], list[_R6], list[_R7]
]: ...


@overload
def multi_apply[**Ps, R1, R2, R3, R4, R5, R6](
    func: Callable[_Ps, tuple[_R1, _R2, _R3, _R4, _R5, _R6]],
    *args: Iterable[Any],
    **kwargs: Any,
) -> tuple[list[_R1], list[_R2], list[_R3], list[_R4], list[_R5], list[_R6]]: ...


@overload
def multi_apply[**Ps, R1, R2, R3, R4, R5](
    func: Callable[_Ps, tuple[_R1, _R2, _R3, _R4, _R5]],
    *args: Iterable[Any],
    **kwargs: Any,
) -> tuple[list[_R1], list[_R2], list[_R3], list[_R4], list[_R5]]: ...


@overload
def multi_apply[**Ps, R1, R2, R3, R4](
    func: Callable[_Ps, tuple[_R1, _R2, _R3, _R4]],
    *args: Iterable[Any],
    **kwargs: Any,
) -> tuple[list[_R1], list[_R2], list[_R3], list[_R4]]: ...


@overload
def multi_apply[**Ps, R1, R2, R3](
    func: Callable[_Ps, tuple[_R1, _R2, _R3]],
    *args: Iterable[Any],
    **kwargs: Any,
) -> tuple[list[_R1], list[_R2], list[_R3]]: ...


@overload
def multi_apply[**Ps, R1, R2](
    func: Callable[_Ps, tuple[_R1, _R2]],
    *args: Iterable[Any],
    **kwargs: Any,
) -> tuple[list[_R1], list[_R2]]: ...


@overload
def multi_apply[**Ps, R1](
    func: Callable[_Ps, _R1],
    *args: Iterable[Any],
    **kwargs: Any,
) -> tuple[list[_R1]]: ...


# @overload
# def multi_apply(
#     func: Callable[_Ps, tuple[Unpack[_RM]]],
#     *args: Iterable[Any],
#     **kwargs: Any,
# ) -> tuple[Unpack[_RM]]:  # technically not correct; each item will be cast to a list.
#     ...


def multi_apply[**Ps](
    func: Callable[_Ps, Any],
    *args: Iterable[Any],
    **kwargs: Any,
) -> tuple[list[Any], ...]:
    """
    Maps each argument iterable to a function, then transposes the results.
    Keyword arguments can be provided as parameters to all functions.

    Parameters
    ----------
    func : Callable.
        Any callable.
    *args: Iterable[...]
        Mapped parameters.
    **kwargs:
        Constant parameters.

    Returns
    -------
    Transposed mapping
    """
    pfunc = partial(func, **kwargs) if kwargs else func

    res_lon = map(pfunc, *args)
    res_zip = zip(*res_lon, strict=False)
    return tuple(map(list, res_zip))  # type: ignore


def multi_merge(dicts: Sequence[dict[str, Tensor]], dim: int):
    """
    Merge a list of dictionaries with tensors along a given dimension.
    All items in the list must have the same keys.

    Parameters
    ----------
    dicts : Sequence[Dict[str, Tensor]]
        List of dictionaries to merge.

    Returns
    -------
    Dict[str, Tensor]
        Merged dictionary.
    """
    keys = dicts[0].keys()

    return {k: torch.cat([d[k] for d in dicts], dim=dim) for k in keys}


_Pt = ParamSpec("_Pt")
_Rt = TypeVar("_Rt", covariant=True)


def traceable[**Pt, Rt](f: Callable[_Pt, _Rt]) -> Callable[_Pt, _Rt]:
    f = cast(Callable[Pt, Rt], allow_in_graph(f))

    @wraps(f)
    def wrapper(*args: _Pt.args, **kwargs: _Pt.kwargs):
        return f(*args, **kwargs)

    return wrapper


_T = TypeVar("_T", int, float, str, torch.Tensor, bool)


@overload
def to_ntuple(n: Literal[1]) -> Callable[[_T | Iterable[_T]], tuple[_T]]: ...


@overload
def to_ntuple(n: Literal[2]) -> Callable[[_T | Iterable[_T]], tuple[_T, _T]]: ...


@overload
def to_ntuple(n: Literal[3]) -> Callable[[_T | Iterable[_T]], tuple[_T, _T, _T]]: ...


@overload
def to_ntuple(
    n: Literal[4],
) -> Callable[[_T | Iterable[_T]], tuple[_T, _T, _T, _T]]: ...


def to_ntuple(n: int) -> Callable[[_T | Iterable[_T]], tuple[_T, ...]]:
    """Ensure that the input, which is optionally an iterable, is casted to a tuple of length `n`."""

    def parse(x: _T | Iterable[_T]) -> tuple[_T, ...]:
        res: tuple[_T, ...]
        if (
            isinstance(x, collections.abc.Iterable) or inspect.isgenerator(x)
        ) and not isinstance(x, str):
            res = tuple(v for v in x)[:n]
        else:
            res = tuple(repeat(x, n))  # type: ignore

        assert isinstance(res, tuple), f"Expect tuple, but got {type(res)}"
        assert len(res) == n, f"Expect tuple of length {n}, but got {res}"

        return res

    return parse


to_1tuple = to_ntuple(1)
to_2tuple = to_ntuple(2)
to_3tuple = to_ntuple(3)
to_4tuple = to_ntuple(4)
to_ntuple = to_ntuple


def to_sequence(x: _T | Sequence[_T]) -> Sequence[_T]:
    """Ensure that the input is casted to a tuple when it is not yet a sequence."""
    if isinstance(x, collections.abc.Sequence) and not isinstance(x, str):
        return x
    return (x,)  # type: ignore


_Td = TypeVar("_Td", float, torch.Tensor)


def make_divisible(
    v: _Td, divisor: int = 8, min_value: int | None = None, round_limit: _Td = 0.9
) -> _Td:
    """Ensure that the input is divisible by `divisor`."""
    min_value = min_value or divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    if new_v < round_limit * v:
        new_v += divisor
    return new_v  # type: ignore


def extend_tuple(x: tuple[_T] | list[_T] | _T, n: int) -> tuple[_T, ...]:
    x = (x,) if not isinstance(x, (tuple, list)) else tuple(x)
    pad_n = n - len(x)
    if pad_n <= 0:
        return x[:n]
    return x + (x[-1],) * pad_n



"""Dynamic convolution operation."""

import torch


def _dynconv2d_naive(
    feature: torch.Tensor,  # (B, C, H, W)
    kernels: torch.Tensor,  # (B, K, C)
) -> torch.Tensor:
    batch_size = feature.shape[0]

    feature = feature.flatten(0, 1)  # (B*C, H, W)
    feature = feature.unsqueeze(0)  # (1, B*C, H, W)

    kernels = kernels.unflatten(-1, (-1, 1, 1))  # (B, K, C, 1, 1)
    kernels = kernels.flatten(0, 1)  # (B*K, C, 1, 1)

    results = torch.nn.functional.conv2d(
        feature, kernels, stride=1, padding=0, groups=batch_size
    )  # (1, B*K, H, W)
    results = results.squeeze(0).unflatten(0, (batch_size, -1))  # (B, K, H, W)
    return results.contiguous()


def _dynconv2d_einsum(
    feature: torch.Tensor,  # (B, C, H, W)
    kernels: torch.Tensor,  # (B, K, C)
) -> torch.Tensor:
    return torch.einsum("bkc, bchw -> bkhw", kernels, feature).contiguous()


def _dynconv2d_matmul(
    feature: torch.Tensor,  # (B, C, H, W)
    kernels: torch.Tensor,  # (B, K, C)
) -> torch.Tensor:
    # (K, C) x (C, H*W) -> (K, H*W)
    return (
        torch.bmm(kernels, feature.flatten(2))
        .unflatten(-1, feature.shape[2:])
        .contiguous()
    )


def dynconv2d(feature: torch.Tensor, kernels: torch.Tensor) -> torch.Tensor:
    """Perform dynamic convolution on the features using the flattened kernels."""
    # Use the fastest implementation available
    return _dynconv2d_matmul(feature, kernels)



"""Layers that can generalically be used in a neural network."""

from . import (
    activations,
    conv,
    drop,
    init,
    linear,
    losses,
    norms,
    posenc,
    projection,
    scale,
    smooth,
    squeeze_excite,
    transformer,
    wrappers,
)
from ._args import *
from ._interpolate import *
from ._sequential import *



"""Implements segmentation-to-depth and vice-versa losses."""

from __future__ import annotations

import typing as T
from typing import override

import torch
from einops import rearrange
from torch import Tensor, nn

from unipercept.nn.losses.functional import split_into_patches
from unipercept.nn.losses.mixins import ScaledLossMixin, StableLossMixin


def _abs_to_norm(
    value: Tensor,
    min_depth: float | Tensor,
    max_depth: float | Tensor,
    mode: str,
) -> Tensor:
    if mode == "absolute":
        result = (value - min_depth) / (max_depth - min_depth)
    elif mode == "disparity":
        min_disp = 1 / max_depth
        max_disp = 1 / min_depth
        scaled_disp = 1 / value
        result = (scaled_disp - min_disp) / (max_disp - min_disp)
    else:
        msg = f"Invalid prediction mode: {mode}"
        raise NotImplementedError(msg)
    return torch.nan_to_num(result, nan=0, posinf=0, neginf=0)


def depth_guided_segmentation_loss(
    seg_feat: Tensor,
    dep_true: Tensor,
    tau: float,
    patch_size: tuple[int, int],
    patch_stride: tuple[int, int],
    valid_ratio: float = 0.5,
    eps: float = 1e-8,
) -> Tensor:
    r"""
    Compute the depth-guided segmentation loss.

    Parameters
    ----------
    seg_feat : Tensor[B, C, H, W]
        The segmentation features. No normalization is applied, depending on the
        initialization of the feature space, it could be beneficial to normalize
        the features in the channel dimension before passing them to this function.
    dep_true : Tensor[B, H, W]
        The ground truth depth map. No normalization is applied, it is recommended to
        tune the ``tau`` parameter to the range of the depth values.
    tau : int
        The temperature parameter for the depth loss.
    patch_size : Tuple[int, int]
        The size of the patch to use. If the size of the depth map is larger than
        the segmentation map, then the patch size will be scaled accordingly and
        median pooling is applied such that the depth map matches the segmentation map.
    patch_stride : Tuple[int, int]
        The stride of the patch.
    valid_ratio : float
        The ratio of depth values that must be valid within a patch to have it be
        a valid patch.
    eps : float
        A small value to avoid division by zero.
    """
    if dep_true.ndim == 4:
        if dep_true.shape[1] == 1:
            dep_true = dep_true.squeeze(1)
        else:
            msg = f"Expected B x 1 x H x W or B x H x W, got: {dep_true.shape}"
            raise ValueError(msg)
    assert dep_true.ndim == 3, f"Expected B x H x W, got: {dep_true.shape}"
    assert seg_feat.ndim == 4, f"Expected B x C x H x W, got: {seg_feat.shape}"

    # Ground truth processing
    with torch.no_grad():
        # Compute the the scale of the patch size and stride such that an equal amount
        # of patches are generated
        W_seg, H_seg = seg_feat.shape[-2:]
        W_dep, H_dep = dep_true.shape[-2:]
        scale_x = W_dep // W_seg
        scale_y = H_dep // H_seg
        dep_patch = split_into_patches(
            dep_true,
            (int(patch_size[0] * scale_x), int(patch_size[1] * scale_y)),
            (int(patch_stride[0] * scale_x), int(patch_stride[1] * scale_y)),
        )

        # Downsample patches via median pooling to match the segmentation size
        dep_patch = rearrange(
            dep_patch,
            "n p (h pH) (w pW) -> n p h w (pH pW)",
            h=patch_size[0],
            w=patch_size[1],
        )
        dep_valid = dep_patch > eps
        dep_patch[~dep_valid] = torch.nan
        dep_patch = torch.nanmedian(dep_patch, dim=-1).values
        dep_valid = dep_valid.int().sum(dim=-1) / dep_valid.shape[-1] > valid_ratio
        dep_patch[~dep_valid] = 0.0

        c_x = patch_size[0] // 2
        c_y = patch_size[1] // 2
        center_mask = torch.zeros(patch_size, device=dep_patch.device, dtype=torch.bool)
        center_mask[c_x, c_y] = True

        dep_anchor = dep_patch[..., center_mask]
        dep_patch = dep_patch[..., ~center_mask]
        dep_diff = (dep_anchor - dep_patch).abs()

        valid_diff = (dep_patch > 0) & (dep_anchor > 0)  # (B, P, pH * pW)
        valid_amount = valid_diff.float().sum(dim=(-2, -1))  # (B)

    # if not valid_amount.any():
    #    return seg_feat * 0.0

    # Segmentation features
    seg_patch = split_into_patches(seg_feat, patch_size, patch_stride)
    seg_center = seg_patch[..., center_mask]
    seg_patch = seg_patch[..., ~center_mask]
    seg_diff = torch.linalg.vector_norm((seg_center - seg_patch), dim=1, ord=2.0)

    # seg_center = seg_patch[..., c_x : c_x + 1, c_y : c_y + 1]  # .contiguous()
    # seg_diff = torch.norm(seg_center - seg_patch, dim=1)  # .clamp(min=eps)

    # valid_diff = valid_diff & (seg_diff > eps)

    loss = torch.exp(-dep_diff / tau) * torch.exp(-(seg_diff**2))  # (B, P, pH * pW)
    loss = loss * valid_diff.float()
    return loss.sum(dim=(-2, -1)) / valid_amount.clamp(min=1)  # (B)


class DGPLoss(StableLossMixin, ScaledLossMixin, nn.Module):
    """Implements a depth-guided panoptic loss (DGP) loss."""

    tau: T.Final[float]
    min_depth: T.Final[float]
    max_depth: T.Final[float]
    mode: T.Final[str]
    patch_size: T.Final[tuple[int, int]]
    patch_stride: T.Final[tuple[int, int]]

    def __init__(
        self,
        *,
        tau=1.0,
        min_depth: float = 1.0,
        max_depth: float = 100.0,
        mode: T.Literal["absolute", "disparity"] = "disparity",
        patch_size: tuple[int, int] = (5, 5),
        patch_stride: tuple[int, int] = (1, 1),
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.tau = tau
        self.patch_size = patch_size
        self.patch_stride = patch_stride
        self.min_depth = min_depth
        self.max_depth = max_depth
        self.mode = mode

    @override
    def forward(self, seg_feat: Tensor, dep_true: Tensor):
        # Normalize depth values to the range [0, 1]
        with torch.no_grad():
            dep_true = _abs_to_norm(dep_true, self.min_depth, self.max_depth, self.mode)

        # Normalize each channel of the segmentation feature space
        seg_feat = nn.functional.normalize(seg_feat, dim=1, p=2.0)

        # Compute loss
        loss = depth_guided_segmentation_loss(
            seg_feat,
            dep_true,
            self.tau,
            self.patch_size,
            self.patch_stride,
            self.eps,
        )
        return loss.mean() * self.scale


def segmentation_guided_triplet_loss(
    dep_feat: Tensor,
    seg_true: Tensor,
    margin: float,
    threshold: int,
    patch_size: tuple[int, int],
    patch_stride: tuple[int, int],
) -> Tensor:
    if seg_true.ndim == dep_feat.ndim - 1:
        seg_true = seg_true.unsqueeze(1)

    assert dep_feat.ndim == 4, (
        f"Expected features as B x C x H x W, got: {dep_feat.shape} for segmentation {dep_feat.shape}"
    )
    assert seg_true.ndim == dep_feat.ndim, (
        f"Expected segmentatino as B x 1 x H x W, got: {seg_true.shape} for depth features {dep_feat.shape}!"
    )

    with torch.no_grad():
        seg_true = nn.functional.interpolate(
            seg_true.float(),
            size=dep_feat.shape[-2:],
            mode="nearest-exact",
        )
        seg_patch = split_into_patches(
            seg_true, patch_size, patch_stride
        )  # B x 1 x P x 5 x 5

        c_x = patch_size[0] // 2
        c_y = patch_size[1] // 2
        center_mask = torch.zeros(patch_size, device=seg_patch.device, dtype=torch.bool)
        center_mask[c_x, c_y] = True

        tgt_anchor = seg_patch[..., center_mask]  # (B, 1, P, 1)
        tgt_values = seg_patch[..., ~center_mask]  # (B, 1, P, pH * pW - 1 = N)

        mask_same = (tgt_anchor == tgt_values).any(dim=-1, keepdim=True)
        mask_anchor = tgt_anchor >= 0
        mask_values = tgt_values >= 0
        mask_valid = (mask_same & mask_anchor).expand_as(mask_values) & mask_values

        w_pos = ((tgt_values == tgt_anchor) & mask_valid).float()
        n_pos = w_pos.sum(dim=-1)

        w_neg = ((tgt_values != tgt_anchor) & mask_valid).float()
        n_neg = w_neg.sum(dim=-1)

        mask_threshold = (n_pos > threshold) & (n_neg > threshold)  # (B, 1, P)
        w_pos[~mask_threshold].zero_()
        w_neg[~mask_threshold].zero_()

    # Split depth features into patches
    dep_patch = split_into_patches(
        dep_feat, patch_size, patch_stride
    )  # B x C x P x 5 x 5
    src_anchor = dep_patch[..., center_mask]  # (B, C, P, 1)
    src_values = dep_patch[..., ~center_mask]  # (B, C, P, N)

    # Expand the anchors such that they have the same shape as the values
    # src_anchor = src_anchor.expand_as(src_values)  # (B, C, P, N)
    src_pos = src_values * w_pos  # (N, C, P, N)
    src_neg = src_values * w_neg

    # Compute the distance (l2) between the anchor and the positive/negative features
    def _distance_to_anchor(a: Tensor, b: Tensor) -> Tensor:
        d = a - b  # a * w - b  # (B, C, P, N)
        return torch.linalg.vector_norm(d, dim=1, ord=2.0, keepdim=True)  # (B, 1, P, N)
        # d = d.sum(dim=-1) / w.sum(dim=-1).clamp_min(1.0)  # (B, 1, P)

    d_pos = _distance_to_anchor(src_anchor, src_pos)
    d_neg = _distance_to_anchor(src_anchor, src_neg)
    # d_neg = torch.norm(
    #    src_anchor * weights_neg - src_neg, p=2, dim=-1
    # ) / n_neg.clamp_min(1.0)

    # Total loss for all patches
    # loss = d_pos + margin - d_neg  # (B, 1, P, N)
    loss = ((d_pos - d_neg).abs() - margin).relu()  # (B, 1, P)

    # Take the weighted average per batch item, then the mean over all batches
    w_total = w_pos + w_neg
    return loss.flatten(1).sum(dim=-1) / w_total.flatten(1).sum(dim=-1).clamp_min(1.0)


class PGTLoss(StableLossMixin, ScaledLossMixin, nn.Module):
    """
    Panoptic-guided Triplet Loss (PGT) loss.

    Paper: https://arxiv.org/abs/2210.07577
    """

    patch_size: T.Final[tuple[int, int]]
    patch_stride: T.Final[tuple[int, int]]
    margin: T.Final[float]
    threshold: T.Final[int]

    def __init__(
        self,
        *,
        patch_size: tuple[int, int] = (5, 5),
        patch_stride: tuple[int, int] | None = (1, 1),
        margin=1e-2,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.patch_size = patch_size
        self.patch_stride = patch_stride or patch_size
        self.margin = margin
        self.threshold = max(1, min(*patch_size) // 2)

    @override
    def forward(self, dep_feat: Tensor, seg_true: Tensor):
        dep_feat = nn.functional.normalize(dep_feat, dim=1, p=2.0)
        loss = segmentation_guided_triplet_loss(
            dep_feat,
            seg_true,
            self.margin,
            self.threshold,
            self.patch_size,
            self.patch_stride,
        )
        return loss.mean() * self.scale


class PGSLoss(ScaledLossMixin, nn.Module):
    """
    Panoptic-guided Smoothness Loss (PGS).

    Paper: https://arxiv.org/abs/2210.07577
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    @override
    def forward(self, disparity: Tensor, panoptic: Tensor):
        # Compute the Iverson bracket for adjacent pixels along the x-dimension
        panoptic_diff_x = (torch.diff(panoptic, dim=-1) != 0).int()

        # Compute the Iverson bracket for adjacent pixels along the y-dimension
        panoptic_diff_y = (torch.diff(panoptic, dim=-2) != 0).int()

        # Compute the partial disp derivative along the x-axis
        disp_diff_x = torch.diff(disparity, dim=-1)

        # Compute the partial disp derivative along the y-axis
        disp_diff_y = torch.diff(disparity, dim=-2)

        loss_x = torch.mean(torch.abs(disp_diff_x) * (1 - panoptic_diff_x))
        loss_y = torch.mean(torch.abs(disp_diff_y) * (1 - panoptic_diff_y))

        return self.scale * (loss_x + loss_y)
